"""Base classes for Waylay REST Services and Resources."""
__docformat__ = "google"

from typing import (
    Optional, Type, TypeVar,
    Mapping, List, Dict, Optional,
    Callable, Collection, Any, Union,
    cast
)
try:
    from typing import Protocol
except ImportError:
    # typing.Protocol is a 3.8 feature ...
    # ... but typing_extensions provides forward compatibility.
    from typing_extensions import Protocol  # type: ignore

from string import Template

from simple_rest_client.api import Resource, API
from waylay.config import WaylayConfig
from waylay.exceptions import ConfigError

S = TypeVar('S', bound='WaylayService')
RS = TypeVar('RS', bound='WaylayRESTService')
R = TypeVar('R', bound='WaylayResource')


class WaylayServiceContext(Protocol):
    """View protocol for the dynamic service context."""

    def get(self, service_class: Type[S]) -> Optional[S]:
        """Get the service instance for the given class, if available."""

    def require(self, service_class: Type[S]) -> S:
        """Get the service instances for the given class or raise a ConfigError."""

    def list(self) -> List['WaylayService']:
        """List all available service instances."""


DEFAULT_SERVICE_TIMEOUT = 10


def _arg_doc(arg):
    desc = f"    {arg['name']} ({arg['type']}): {arg['description']}"
    if 'examples' not in arg:
        return desc
    examples = '\n'.join(
        f'        - {ex}' for ex in arg['examples']
    )
    return (
        desc +
        f'\n       Examples:\n{examples}'
    )

# Implementation Note
# -------------------
# Current solution is scaffolded around the `simple_rest_client`, but is planned to be refactored.


class WaylayAction(dict):
    """Configuration object for a service action."""

    @property
    def resource(self) -> 'WaylayResource':
        """Get the parent resource of this action."""
        return self['_resource']

    @property
    def id(self) -> str:
        """Get the action id."""
        return self['id']

    @property
    def name(self) -> str:
        """Get the action name as exposed in the SDK."""
        return self.get('name', self.id)

    @property
    def description(self) -> Optional[str]:
        """Get the action description."""
        desc = self.get('description')
        if desc:
            return desc
        return (self.action_method.__doc__ or '').split("\n")[0]

    @property
    def arguments(self) -> List[Dict]:
        """Get the arguments documentation info."""
        return self.get('arguments', [])

    @property
    def returns(self) -> List[Dict]:
        """Get the returns documentation info."""
        return self.get('returns', [])

    @property
    def doc_links(self) -> Dict[str, Dict[str, str]]:
        """Get (documentation) links templates."""
        roots = self.resource.hal_roots
        return {
            link: dict(href=roots.get(link, '') + ref)
            for link, ref in self.get('links', {}).items()
        }

    @property
    def rest_action_doc(self) -> str:
        """Get the documentation string for the REST action."""
        return ''

    @property
    def action_method(self) -> Callable:
        """Get the SDK action method that is used to expose this action."""
        return getattr(self.resource, self.id)

    @property
    def sdk_action_method(self) -> Callable:
        """Get the REST action method that is used to expose this action."""
        return getattr(self.resource, self.name)

    @property
    def sdk_action_doc(self) -> str:
        """Get the sdk python documentation for this action."""
        doc = self.sdk_action_method.__doc__ or ''
        if self.description and doc.startswith(self.description):
            doc = doc[len(self.description):]
        if self.name != self.id:
            # wrapped actions
            doc += "\n" + (self.action_method.__doc__ or '')
        return doc

    @property
    def arguments_doc(self) -> str:
        """Get the argument documentation for this action."""
        args = self.arguments
        if not args:
            return ''
        return '  Arguments:\n' + '\n'.join(
            _arg_doc(arg)
            for arg in args
        )

    @property
    def returns_doc(self) -> str:
        """Get the return value documentation for this action."""
        returns = self.returns
        if not returns:
            return ''
        return '  Returns: \n' + '\n'.join(
            _arg_doc(ret)
            for ret in returns
        )

    def __repr__(self):
        """Get a string representation of this action."""
        return self.name

    def as_dict(self):
        """Get a dictionary representation."""
        return {
            'name': self.name,
            'description': self.description
        }

    @property
    def fqn(self):
        """Get the fully qualified name for this action."""
        return f'{self.resource.fqn}.{self.name}'


class WaylayRESTAction(WaylayAction):
    """Configuration object representing a single Waylay REST action."""

    @property
    def method(self) -> Optional[str]:
        """Get the action HTTP method."""
        return self.get('method')

    @property
    def url(self) -> Optional[str]:
        """Get the action url template."""
        return self.get('url')

    def __repr__(self):
        """Get a string representation of this REST action."""
        return f"{self.method} {self.url}"

    @property
    def rest_action_doc(self) -> str:
        """Get the documentation string for the REST action."""
        url_interpolated = self.url
        if self.arguments and self.url:
            url_interpolated = self.url.format(*[f"{{{arg['name']}}}" for arg in self.arguments])
        return f'{self.method} {self.resource.service.gateway_root_path}{url_interpolated}'

    def as_dict(self):
        """Get a dictionary representation."""
        return {
            'name': self.name,
            'description': self.description,
            'rest': self.rest_action_doc
        }


class WaylayRESTActionsWrapper(WaylayAction):
    """Configuration object representing a method that wraps one or more REST actions."""

    @property
    def wrapped_actions(self) -> List[WaylayRESTAction]:
        """Get the list of wrapped REST actions as defined by 'wrapped_actions'."""
        return [
            cast(WaylayRESTAction, self.resource.actions[id])
            for id in self.get('wrapped_actions', [])
        ]

    @property
    def doc_links(self) -> Dict[str, Dict[str, str]]:
        """Get (documentation) links templates."""
        roots = self.resource.hal_roots
        return {
            link: dict(href=roots.get(link, '') + ref)
            for wrapped_action in self.wrapped_actions
            for link, ref in wrapped_action.get('links', {}).items()
        }


def create_waylay_action(**kwargs) -> WaylayAction:
    """Initialize a waylay action that wraps a REST call or other method."""
    if 'url' in kwargs:
        return WaylayRESTAction(**kwargs)
    if 'wrapped_actions' in kwargs:
        return WaylayRESTActionsWrapper(**kwargs)
    raise ValueError("A waylay action configuration requires either an 'url' or 'wrapped_actions'.")


class WaylayResource(Resource):
    """Client object representing a Waylay REST Resource.

    This is a collection of REST operations that have a single Waylay Entity as subject.
    """

    service: 'WaylayService'
    resource_name: str
    actions: Dict[str, Union[WaylayAction, Collection[Any]]]

    # class variables
    link_roots: Dict[str, str] = {}

    def __init__(self, *args, **kwargs):
        """Create a Waylay Resource."""
        self.service = kwargs.pop('service', None)
        self.actions = {
            id: create_waylay_action(id=id, _resource=self, **action)
            for id, action in self.actions.items()
        }
        super().__init__(*args, **kwargs)

    @property
    def name(self):
        """Get the name that identifies this resource in the Python SDK."""
        return self.resource_name

    @property
    def description(self):
        """Get a description of this service."""
        return self.__doc__

    def add_action(self, action_name: str):
        """Add action, and apply decorators."""
        if isinstance(self.actions[action_name], WaylayRESTAction):
            super().add_action(action_name)
        self.decorate_action(action_name)

    def decorate_action(self, action_name):
        """Decorate the action if a 'decorators' definition exist."""
        action = self.get_action(action_name)
        decorators = action.get('decorators', None)
        action_method = getattr(self, action_name)
        action_method.__dict__['action'] = action
        if decorators:
            for decorator in decorators:
                action_method = decorator(action_method)
                action_method.__dict__['action'] = action
        setattr(self, action_name, action_method)

    def __repr__(self):
        """Get a technical string representation of this action."""
        actions_repr = ', '.join(
            f"{name}: {action_def}"
            for name, action_def in self.actions.items()
        )
        return (
            f"<{self.__class__.__name__}("
            f"actions=[{actions_repr}]"
            ")>"
        )

    def as_dict(self):
        """Get a dictionary representation."""
        return {
            'name': self.resource_name,
            'description': self.description,
            'actions': [
                a.as_dict()
                for a in self.actions.values()
            ],
            '_links': self.doc_links(None)
        }

    @property
    def hal_roots(self) -> Dict[str, str]:
        """Get the root urls for documentation links if this resource."""
        return {
            link: Template(
                href_root
            ).safe_substitute(
                root_url=self.api_root_url,
                doc_url=self.service.config.doc_url,
                apidoc_url=self.service.config.apidoc_url,
            )
            for link, href_root in self.link_roots.items()
        }

    def doc_links(self, action: Optional[str]) -> Dict[str, Dict[str, str]]:
        """Create a HAL `_links` representation for (documentation) links.

        Arguments:
            action      if specified, give links for a specific action rather than the resource.
        """
        hrefs = {rel: '' for rel in self.link_roots}
        if action:
            hrefs = cast(WaylayAction, self.actions[action])['links']

        return {
            link_rel: dict(
                href=Template(
                    href_root + hrefs.get(link_rel, '')
                ).safe_substitute(
                    root_url=self.api_root_url,
                    doc_url=self.service.config.doc_url,
                    apidoc_url=self.service.config.apidoc_url,
                )
            )
            for link_rel, href_root in self.link_roots.items()
            if link_rel in hrefs
        }

    # override
    def get_action_full_url(self, action_name, *parts):
        """Override the regular url computation when not using api gateway."""
        if (self.api_root_url is None):
            srv = self.service
            raise ConfigError(
                f'The service `{srv.service_key}` has no url configuration. '
                f'Please provide a endpoint using a setting with key `{srv.config_key}`,'
                'or request your tenant administrator '
                f'to configure the global setting `waylay_{srv.config_key}`.'
            )
        return super().get_action_full_url(action_name, *parts)

    @property
    def fqn(self):
        """Get the fully qualified name for this resource."""
        return f"{self.service.fqn if self.service else '_'}.{self.name}"


class WaylayService():
    """Client object representing a Waylay Tool."""

    resource_definitions: Mapping[str, Type[Resource]]
    config: WaylayConfig
    service_key: str = ''
    plugin_priority = 0
    gateway_root_path: Optional[str] = None
    _resources: Dict[str, WaylayResource]

    def __init__(self, *args, **kwargs):
        """Create a WaylayRESTService."""
        self._resources = getattr(self, '_resources', {})
        for name, resource_class in self.resource_definitions.items():
            self._add_waylay_resource(resource_name=name, resource_class=resource_class)

    @property
    def name(self):
        """Get the name that identifies this service in the Python SDK."""
        return self.service_key

    @property
    def description(self):
        """Get a description of this service."""
        return self.__doc__

    @property
    def root_url(self) -> Optional[str]:
        """Get the root url."""
        return self.get_root_url()

    @property
    def resources(self):
        """Get the resources supported by this service."""
        return self.list_resources()

    def _add_waylay_resource(self, resource_name: str, resource_class: Type[R], **kwargs) -> R:
        waylay_resource = resource_class(service=self)
        setattr(self, resource_name, waylay_resource)
        self._resources[resource_name] = waylay_resource
        waylay_resource.resource_name = resource_name
        waylay_resource.service = self
        return waylay_resource

    def list_resources(self) -> List[WaylayResource]:
        """List the WaylayResources of this service."""
        return list(self._resources.values())

    def configure(self: S, config: WaylayConfig, context: WaylayServiceContext) -> S:
        """Configure endpoints and authentication with given configuration.

        Returns self
        """
        self.config = config
        return self.reconfigure()

    def reconfigure(self: S) -> S:
        """Configure endpoints and authentication with current configuration.

        Returns self
        """
        return self

    def get_root_url(self) -> Optional[str]:
        """Get the root url."""
        return None

    def __repr__(self):
        """Get a technical string representation of this tool."""
        return (
            f"<{self.__class__.__name__}("
            f"service_key={self.service_key},"
            ")>"
        )

    @property
    def fqn(self):
        """Get the fully qualified name for this service."""
        return f'waylay:{self.service_key}'


class WaylayRESTService(API, WaylayService):
    """Client object representing a Waylay Service.

    A collection of Resources with their operations.
    """

    # class variables
    config_key: str = 'api'
    default_root_path: str = ''
    link_templates: Dict[str, str] = {}

    def __init__(self, *args, **kwargs):
        """Create a WaylayRESTService."""
        timeout = kwargs.pop('timeout', DEFAULT_SERVICE_TIMEOUT)
        json_encode_body = kwargs.pop('json_encode_body', True)
        API.__init__(self, *args, timeout=timeout, json_encode_body=json_encode_body, **kwargs)
        WaylayService.__init__(self, *args, **kwargs)

    def _add_waylay_resource(self, resource_name: str, resource_class: Type[R], **kwargs) -> R:
        self.add_resource(resource_name=resource_name, resource_class=resource_class, **kwargs)
        waylay_resource: R = self._resources[self.correct_attribute_name(resource_name)]
        waylay_resource.service = self
        return waylay_resource

    def set_root_url(self, root_url):
        """Set the root url and reconfigure the service."""
        self.config.set_root_url(self.config_key, root_url)
        self.reconfigure()

    def get_root_url(self) -> Optional[str]:
        """Get the root url."""
        if self.config is None:
            return None
        return self.config.get_root_url(
            self.config_key,
            gateway_root_path=self.gateway_root_path,
            default_root_path=self.default_root_path
        )

    def reconfigure(self: RS) -> RS:
        """Configure endpoints and authentication with current configuration.

        Returns self
        """
        if self.config is None:
            return self
        root_url = self.get_root_url()
        for resource in self._resources.values():
            resource.api_root_url = root_url
            resource.client.auth = self.config.auth
        return self

    def __repr__(self):
        """Get a technical string representation of this service."""
        return (
            f"<{self.__class__.__name__}("
            f"service_key={self.service_key},"
            f"config_key={self.config_key},"
            f"root_url={self.get_root_url()},"
            f"resources=[{', '.join(self._resources.keys())}]"
            ")>"
        )

    def doc_links(self) -> Dict[str, Dict[str, str]]:
        """Create a HAL `_links` representation for the documentation links."""
        return {
            rel: dict(href=Template(href).safe_substitute(
                root_url=self.root_url,
                doc_url=self.config.doc_url,
                apidoc_url=self.config.apidoc_url,
            ))
            for rel, href in self.link_templates.items()
        }



"""Decorators to be used in waylaycli."""
import logging

from functools import wraps
from waylay.exceptions import WaylayError

LOG = logging.getLogger(__name__)


def cli_exeption_decorator(action_method):
    """Create a decorator that handles exceptions for the cli tool."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except BrokenPipeError as exc:
            LOG.debug('Ignored broken pipe (stdout gone)', exc_info=True)
        except (WaylayError, ValueError, TypeError) as exc:
            LOG.error('\n'.join(exc.args), exc_info=LOG.isEnabledFor(logging.DEBUG))

    return wrapped



"""Waylay Rest Services."""

from ._base import WaylayRESTService, WaylayResource, WaylayServiceContext, WaylayService, WaylayAction
from . import _decorators as decorators

from .byoml import ByomlService
from .timeseries import TimeSeriesService
from .resources import ResourcesService
from .storage import StorageService
from .util import UtilService
from .etl import ETLService
from .data import DataService
from .queries import QueriesService


SERVICES = [
    ByomlService,
    TimeSeriesService,
    QueriesService,
    ResourcesService,
    StorageService,
    UtilService,
    ETLService,
    DataService
]



"""Model objects for the timeseries tool."""
__docformat__ = "google"

from tarfile import TarFile
from typing import (
    Optional, Sequence, Union,
    Tuple, Iterator, Iterable, Any,
    Mapping, SupportsInt,
    Callable, IO, List
)

try:
    from typing import Protocol, runtime_checkable
except ImportError:  # pragma: no cover
    # typing.Protocol is a 3.8 feature ...
    # ... but typing_extensions provides forward compatibility.
    from typing_extensions import Protocol, runtime_checkable  # type: ignore

from datetime import datetime, timedelta, tzinfo
from enum import Enum
import os
import io
import re
import sys
import tempfile
from pathlib import Path

from dataclasses import dataclass, field, asdict, replace
from zipfile import ZipFile
import pandas as pd
import isodate
import pytz

from waylay.exceptions import RequestError

if sys.version_info > (3, 7):
    Pattern = re.Pattern  # pylint: disable=no-member
else:
    Pattern = Any

ETL_IMPORT_BUCKET = 'etl-import'
ETL_IMPORT_UPLOAD_PREFIX = 'upload/'
METRIC_PREFIX = 'waylay.resourcemessage.metric.'
METRIC_PREFIX_LENGTH = len(METRIC_PREFIX)

RESOURCE_COLUMN, METRIC_COLUMN, TIMESTAMP_COLUMN, VALUE_COLUMN = \
    ETL_IMPORT_COLUMN_NAMES = ['resource', 'metric', 'timestamp', 'value']
RESOURCE_IDX, METRIC_IDX, TIMESTAMP_IDX, VALUE_IDX = range(4)

MeasurementValue = Union[str, int, float, bool, None]

Measurement = Tuple[datetime, MeasurementValue]
MeasurementIterator = Iterator[Measurement]
SeriesProvider = Mapping[Tuple[str, str], Iterable[Measurement]]
SeriesIterator = Iterator[Tuple[str, str, MeasurementIterator]]

CSVReader = Iterator[Sequence[str]]
CSVReaderAndResource = Tuple[CSVReader, Optional[str]]
PathLike = Union[str, os.PathLike]

fromisoformat = datetime.fromisoformat

MAX_EPOCH_SECONDS = 9223000000  # '2262-04-07 16:26:40+0000'
MAX_EPOCH_MILLIS = MAX_EPOCH_SECONDS * 1000
MAX_EPOCH_MICROS = MAX_EPOCH_MILLIS * 1000
MAX_EPOCH_NANOS = MAX_EPOCH_MICROS * 1000


def parse_timestamp_unix(epoch: Union[str, SupportsInt]) -> pd.Timestamp:
    """Convert a unix epoch to datetime, automatically recognising ranges.

    Assumes that the represented timestamps are before '2262-04-07 16:26:40+0000'
    and after '1970-04-17 17:56:40+0000' (when not using seconds unit).
    (Limitation from 64-bit nanosecond representation used in pandas).
    """
    int_epoch = int(epoch)
    if int_epoch <= MAX_EPOCH_SECONDS:
        # use float to preserve fractional seconds
        # in smaller units, use int to have predictable rounding
        return pd.Timestamp(float(epoch), unit='s')  # type: ignore
    if int_epoch <= MAX_EPOCH_MILLIS:
        return pd.Timestamp(int_epoch, unit='ms')
    if int_epoch <= MAX_EPOCH_MICROS:
        return pd.Timestamp(int_epoch, unit='us')
    if int_epoch <= MAX_EPOCH_NANOS:
        return pd.Timestamp(int_epoch, unit='ns')
    raise ValueError(f'Not a supported unix epoch: {epoch}')


def _detect_timestamp_unix_unit(epoch: Union[str, SupportsInt]) -> 'TimestampFormat':
    """Detect the unix epoch formatter for  a give example timestamp value."""
    int_epoch = int(epoch)
    if int_epoch <= MAX_EPOCH_SECONDS:
        # use float to preserve fractional seconds
        # in smaller units, use int to have predictable rounding
        return TimestampFormat.SECONDS
    if int_epoch <= MAX_EPOCH_MILLIS:
        return TimestampFormat.MILLIS
    if int_epoch <= MAX_EPOCH_MICROS:
        return TimestampFormat.MICROS
    if int_epoch <= MAX_EPOCH_NANOS:
        return TimestampFormat.NANOS
    raise ValueError(f'Not a supported unix epoch: {epoch}')


def parse_timestamp_auto(value: Union[str, SupportsInt]) -> pd.Timestamp:
    """Convert a value to a timestamp, trying epoch or ISO representations."""
    try:
        return parse_timestamp_unix(value)
    except ValueError:
        return pd.Timestamp(value)


class ParserRequestError(RequestError):
    """Request validation errors in the timeseries etl parser utilities."""


def try_parse_float(value: str) -> Union[float, str]:
    """Parse a string to a float value or return the orignal value."""
    try:
        return float(value)
    except ValueError:
        return value


def parse_float(value: str) -> Optional[float]:
    """Parse a string to a float value or return None."""
    try:
        return float(value)
    except ValueError:
        return None


def try_parse_int(value: str) -> Union[int, str]:
    """Parse a string to a int value or return the orignal value."""
    try:
        return int(float(value))
    except ValueError:
        return value


def parse_int(value: str) -> Optional[int]:
    """Parse a string to a int value or return None."""
    try:
        return int(float(value))
    except ValueError:
        return None


def try_parse_bool(value: str) -> Union[str, bool]:
    """Parse a string to a boolean value or return the orignal value."""
    lower_value = value.lower()
    if lower_value in ['false', 'no', 'n', '0']:
        return False
    if lower_value in ['true', 'yes', 'y', '1']:
        return True
    return value


def parse_bool(value: str) -> Optional[bool]:
    """Parse a string to a boolean value or return None."""
    value = value.lower()
    if value in ['false', 'no', 'n', '0']:
        return False
    if value in ['true', 'yes', 'y', '1']:
        return True
    return None


PARSERS_BY_VALUE_TYPE = {
    'float': parse_float,
    'double': parse_float,
    'string': lambda x: x,
    'bool': parse_bool,
    'integer': parse_int,
}

TIMESTAMP_COLUMN_NAMES = [TIMESTAMP_COLUMN, 'time', 'date']


def render_timestamp_zulu(timestamp: datetime) -> str:
    """Render a timestamp into the format used by the ETL service."""
    return timestamp.isoformat().replace('+00:00', 'Z')


def parse_timestamp(iso_timestamp: str) -> datetime:
    """Parse a timestamp from the format used by the ETL service."""
    return fromisoformat(iso_timestamp.replace('Z', '+00:00'))


def parse_interval(interval: Optional[Union[str, pd.Timedelta, timedelta]]) -> Optional[timedelta]:
    """Parse a str to a time interval, supports ISO8601 durations."""
    if interval is None or isinstance(interval, timedelta):
        return interval

    try:
        if isinstance(interval, str) and 'P' in interval:
            return isodate.parse_duration(interval)
        return pd.Timedelta(interval).to_pytimedelta()

    except ValueError as exc:
        raise ParserRequestError(f'Invalid interval: {exc}') from exc


RESOURCE_METADATA_KEY = dict(
    value_type='valueType',
    metric_type='metricType'
)


@dataclass
class Metric:
    """Metadata for a metric (univariate series) within a dataset.

    This holds both specifications inferred from the dataset, as
    given explicitely by the caller.

    Attributes:
        name                The metric name that should be used in waylay
        key                 The key used in the input data set
        value_parser        The (python) parser for this value when reading from csv, e.g. `float`

    Descriptive Attributes:
        value_type          The (javascript) value_type that for this value
        metric_type         The type of metric (`gauge`, `counter`, `rate`, `timestamp` )
        unit                The unit
        description
    """

    name: str
    key: Optional[str] = None
    value_parser: Optional[Callable[[str], MeasurementValue]] = None

    description: Optional[str] = None
    value_type: Optional[str] = None
    metric_type: Optional[str] = None
    unit: Optional[str] = None

    def __post_init__(self):
        """Remove legacy prefix from name."""
        if self.name.startswith(METRIC_PREFIX):
            if self.key is None:
                self.key = self.name
            self.name = self.name[METRIC_PREFIX_LENGTH:]

    @property
    def key_or_name(self):
        """Get the data key for this metric, defaulting to the metric name."""
        return self.key or self.name

    def get_value_parser(self) -> Callable[[str], MeasurementValue]:
        """Get the parser callback for this metric."""
        if self.value_parser:
            return self.value_parser
        if not self.value_type:
            return try_parse_float
        return PARSERS_BY_VALUE_TYPE.get(self.value_type, try_parse_float)

    def to_dict(self):
        """Convert to a json object representation.

        This format complies with how metric metadata is stored in the Waylay Resource metadata.
        """
        return {
            RESOURCE_METADATA_KEY.get(key, key): value
            for key, value in asdict(self).items()
            if key not in ['key', 'value_parser']
            if value is not None
        }

    def __repr__(self) -> str:
        """Render as string."""
        if self.key:
            return f'{self.name} as {self.key}'
        return self.name


@dataclass
class Resource:
    """Metadata for a resource (owning entity of series) within a dataset.

    This holds both specifications inferred from the dataset,
    as given explicitely by the caller.

    Attributes:
        id:     The resource id that should be used in waylay
        key:    The key used in the input data set

    Descriptive Attributes:
        name:   The resource name that should be used in waylay
        description: A string description of the resource
        metrics:
            Metadata documentation on the series metrics
            that can be uploaded for this resource.
    """

    id: str
    key: Optional[str] = None

    description: Optional[str] = None
    name: Optional[str] = None
    metrics: Optional[Sequence[Metric]] = None

    @property
    def key_or_id(self):
        """Get the data key for this resource, defaulting to the resource id."""
        return self.key or self.id

    def to_dict(self):
        """Convert to a json object representation.

        This format complies with how resource metadata is stored in the Waylay Resource metadata.
        """
        resource_repr = {
            key: value for key, value in asdict(self).items()
            if key not in ['metrics']
            if value is not None
        }
        if 'name' not in resource_repr:
            resource_repr['name'] = self.id
        if self.metrics:
            resource_repr['metrics'] = [m.to_dict() for m in self.metrics]  # pylint: disable=not-an-iterable
        return resource_repr

    def with_metrics(self, metrics: Sequence[Metric]) -> 'Resource':
        """Return a copy of this description object with the given metrics instead."""
        return replace(self, metrics=metrics)

    def __repr__(self) -> str:
        """Render as string."""
        if self.key:
            return f'{self.id} as {self.key}'
        return self.id


class TimestampFormat(str, Enum):
    """Default timestamp formats for input and output."""

    ISO = (
        'iso', 'ISO8601 timestamp',
        '2020-09-17T01:38:59+00:00', r'^iso.*',
        # pd.Timestamp.fromisoformat, ==> cannot parse timezones?
        # fromisoformat # ==> cannot localize to UTC
        pd.Timestamp,  # OK, but parses a lot of other formats
        lambda t: t.isoformat(),
        True
    )
    SECONDS = (
        's', 'Unix seconds epoch', 1631099086, r'^s.*',
        lambda v: pd.Timestamp(float(v), unit='s', tz=pytz.UTC),
        lambda t: int(t.timestamp())
    )
    MILLIS = (
        'ms', 'Unix milliseconds epoch', 1631099086346, r'^(ms|mil).*',
        lambda v: pd.Timestamp(int(v), unit='ms', tz=pytz.UTC),
        lambda t: int(t.timestamp() * 1000)
    )
    MICROS = (
        'us', 'Unix microseconds epoch',  1631099086234123, r'^(u|µ|micr).*',
        lambda v: pd.Timestamp(int(v), unit='us', tz=pytz.UTC),
        lambda t: int(t.timestamp() * 1000000)
    )
    NANOS = (
        'ns', 'Unix nanosecond epoch', 1631099086432532154, r'^(ns|nano).*',
        lambda v: pd.Timestamp(int(v), unit='ns', tz=pytz.UTC),
        lambda t: int(t.timestamp() * 1000000000)
    )
    UNIX = (
        'unix', 'Unix epoch with automatic unit detection. Renders milliseconds.',
        1631099086346, r'^(un|ep).*',
        parse_timestamp_unix,
        lambda t: int(t.timestamp() * 1000)
    )
    AUTO = (
        'auto', 'Parse numeric value as unix epoch or ISO. Render ISO timestamp',
        '2020-09-17T01:38:59+00:00', r'^auto.*',
        parse_timestamp_auto,
        lambda t: t.isoformat(),
    )

    description: str
    example: Any
    _spec_matcher: Pattern
    # wrapping in a tuple to avoid mypy interpreting callable members as methods
    _parse: Tuple[Callable[[Any], datetime]]
    _format: Tuple[Callable[[datetime], Any]]
    _timezone_dependent: bool

    def __new__(
        cls,
        value: str,
        description: str,
        example: Any,
        spec_matcher: str,
        parse_method: Callable[[Any], pd.Timestamp],
        format_method: Callable[[pd.Timestamp], Any],
        timezone_dependent: bool = False,
    ):
        """Construct an enum value.

        See https://docs.python.org/3/library/enum.html#when-to-use-new-vs-init.
        """
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        obj.example = example
        obj._parse = (parse_method,)
        obj._format = (format_method,)
        obj._spec_matcher = re.compile(spec_matcher, re.RegexFlag.IGNORECASE)
        obj._timezone_dependent = timezone_dependent
        return obj

    @classmethod
    def lookup(cls, spec: Union[str, 'TimestampFormat']) -> 'TimestampFormat':
        """Lookup a TimestampFormat from string specification."""
        if isinstance(spec, cls):
            return spec
        for timestamp_format in cls:
            if timestamp_format._spec_matcher.match(spec):
                return timestamp_format
        raise ValueError(f'Invalid timestamp format specification: {spec}')

    @staticmethod
    def _parse_tz(tz: Any = None) -> tzinfo:
        if isinstance(tz, tzinfo):
            return tz
        if tz:
            return pytz.timezone(tz)
        return pytz.UTC

    def parser(self, tz: Any = None) -> Callable[[Any], pd.Timestamp]:
        """Get a parser that uses a (default) timezone while parsing."""
        parser = self._parse[0]
        if not self._timezone_dependent:
            return parser

        timezone = self._parse_tz(tz)

        def _parse_with_tz(value: Any) -> pd.Timestamp:
            ts = parser(value)
            if not ts.tzinfo:
                ts = ts.replace(tzinfo=timezone)
            return ts.astimezone(pytz.UTC)
        return _parse_with_tz

    def formatter(self, tz: Any = None) -> Callable[[pd.Timestamp], Any]:
        """Get a formatter that uses a (default) timezone when relevant."""
        formatter = self._format[0]
        if not self._timezone_dependent:
            return formatter

        timezone = self._parse_tz(tz)
        return lambda ts: formatter(ts.astimezone(timezone))

    def __str__(self):
        """Get the string representation."""
        return self.value


@dataclass
class SeriesSettings:
    """Settings for the mapping of an input (import) or CSV output (export).

    Attributes:
        metric_column:
            The input attribute containing the metric id.
            This is a column name, or a pandas multiindex level name for the columns.
            If not specified:
            * if a `value_column` is specified, the default "metric" column is uses if present.
            * else, each column name (except the `resource_column` and `timestamp_column` ones),
                is a metric key and provide seperate series.
        metric:
            A default metric name for this import.
        metrics:
            A list of either:
            - metric keys retained from the input
            - `waylay.service.timeseries.parser.model.Metric` entries that describe and map metrics
            If specified, only series for these metrics are processed.
        resource_column:
            The input key containing the resource id.
            This is a column name, or a pandas multiindex level name for the columns.
            If not specified, a fixed `resource` should be specified,
            or the column key "resource" is used if present.
        resource:
            A default resource id to use for this import.
        resources:
            A list of either:
            - resource keys retained from the input
            - `waylay.service.timeseries.parser.model.Resource`
                entries that describe and map resources
            If specified, only series for these resources are processed.
        value_column:
            The column key containing the value. When set, the metric names should be
            provided through `metric` or `metric_column` (default "metric").
        timestamp_column:
            The input key containing the timestamp.
        timestamp_offset:
            A time interval to add to the input timestamp.
        timestamp_first:
            Forces the first timestamp, and increments
            the following timestamps with the same amount.
        timestamp_last:
            Forces the last timestamp, and increments the preceding timestamps with the same amount.
        timestamp_interval:
            Ignores the input timestamps and writes data with fixed timestamp intervals.
            Requires `timestamp_first`.
        timestamp_constructor:
            A callable that creates a `datetime` or `pandas.Timestamp` from the timestamp input.
        timestamp_timezone:
            A timezone indicator that should be used to interpret local time data.
        timestamp_from:
            Filters to have a timestamp equal to or greater than this one.
        timestamp_until:
            Filters to have a timestamp strictly less than this one.
    """

    metrics: Optional[Sequence[Union[str, Metric]]] = None
    metric_column: Optional[str] = None
    metric: Optional[str] = None

    resources: Optional[Sequence[Union[str, Resource]]] = None
    resource_column: Optional[str] = None
    resource: Optional[str] = None

    value_column: Optional[str] = None

    timestamp_column: Optional[str] = None
    timestamp_offset: Optional[timedelta] = None
    timestamp_first: Optional[datetime] = None
    timestamp_last: Optional[datetime] = None
    timestamp_interval: Optional[timedelta] = None
    timestamp_constructor: Optional[Callable[[Any], datetime]] = None
    timestamp_timezone: Optional[str] = None

    # TODO: document, review
    name: Optional[str] = None
    timestamp_from: Optional[Any] = None
    timestamp_until: Optional[Any] = None
    timestamp_formatter: Optional[Callable[[datetime], Any]] = None
    timestamp_format: Optional[TimestampFormat] = None
    write_csv_header: bool = True
    per_resource: bool = False
    per_metric: bool = False
    skip_empty_values = True

    def __post_init__(self):
        """Parse string input args to objects."""
        self.timestamp_format = (
            TimestampFormat.lookup(self.timestamp_format) if self.timestamp_format else None
        )
        self.timestamp_offset = parse_interval(self.timestamp_offset)
        self.timestamp_interval = parse_interval(self.timestamp_interval)

    def iter_metrics(self) -> Iterator[Metric]:
        """Iterate the metric specifications if available."""
        if not self.metrics:
            if self.metric:
                yield Metric(name=self.metric)
            return

        for metric in self.metrics:
            if isinstance(metric, Metric):
                metric_spec = metric
            else:
                metric_spec = Metric(name=metric)
            if self.metric and metric_spec.name != self.metric:
                continue
            yield metric_spec

    def iter_resources(self) -> Iterator[Resource]:
        """Iterate the resource specifications if available."""
        if not self.resources:
            if self.resource:
                yield Resource(id=self.resource)
            return

        for resource in self.resources:
            if isinstance(resource, Resource):
                resource_spec = resource
            else:
                resource_spec = Resource(id=resource)
            if self.resource and resource_spec.id != self.resource:
                continue
            yield resource_spec

    def metric_by_key(self, key: Optional[str]) -> Optional[str]:
        """Lookup the actual metric name for the given column key."""
        if not key:
            return self.metric
        if not self.metrics:
            return _remove_prefix(key)
        for metric in self.iter_metrics():
            if _remove_prefix(key) == (_remove_prefix(metric.key) or metric.name):
                return metric.name
        return None

    def key_by_metric(self, name: str) -> Optional[str]:
        """Lookup the column key that is used to represent the a given metric name.

        The empty key resolves to the default metric.
        """
        if not self.metrics:
            return name or self.metric
        for metric in self.iter_metrics():
            if name == metric.name:
                return metric.key or metric.name
        return self.metric

    def resource_by_key(self, key: Optional[str]) -> Optional[str]:
        """Lookup the actual resource id for the given resource key.

        The empty key resolves to the default resource.
        """
        if not key:
            return self.resource
        if not self.resources:
            return key
        for resource in self.iter_resources():
            if key == resource.key_or_id:
                return resource.id
        return self.resource

    def key_by_resource(self, resource_id: str) -> Optional[str]:
        """Lookup the column key that is used to represent the a given resource id."""
        if not self.resources:
            return resource_id
        for resource in self.iter_resources():
            if resource_id == resource.id:
                return resource.key_or_id
        return None

    def has_timestamp(self) -> bool:
        """Check wether the dataset should deliver timestamps."""
        return (
            self.timestamp_column is not None or
            self.timestamp_interval is None or (
                self.timestamp_first is None and
                self.timestamp_last is None
            )
        )

    def metric_for(self, metric_name: str) -> Metric:
        """Create a metadata object for the given metric names.

        Uses the `metrics` registered on this object as a data description catalog.
        """
        return next(
            (m for m in self.iter_metrics() if m.name == metric_name),
            Metric(metric_name)
        )

    def resource_for(
        self, resource_id: str,
        metric_names: Optional[Sequence[str]] = None
    ) -> Resource:
        """Create a metadata object for the given resource id and metric names.

        Uses the `resources` and `metrics` registered on this object as a data
        description catalog to enhance the _resource_ and _metric_
        references found in a concrete series.
        """
        resource_meta = next(
            (r for r in self.iter_resources() if r.id == resource_id),
            Resource(resource_id)
        )
        if metric_names is not None:
            resource_meta = resource_meta.with_metrics(
                list(
                    self.metric_for(metric_name)
                    for metric_name in metric_names
                )
            )
        return resource_meta

    def get_timestamp_formatter(self) -> Callable[[datetime], Any]:
        """Get the formatter for timestamps in csv exports."""
        if self.timestamp_formatter:
            return self.timestamp_formatter
        if self.timestamp_format:
            return self.timestamp_format.formatter(self.timestamp_timezone)
        return TimestampFormat.ISO.formatter(self.timestamp_timezone)

    def get_timestamp_parser(self, example: Any = None) -> Callable[[Any], datetime]:
        """Get the parser for timestamps in csv imports."""
        if self.timestamp_constructor:
            return self.timestamp_constructor
        if self.timestamp_format:
            return self.timestamp_format.parser(self.timestamp_timezone)

        if example:
            try:
                return _detect_timestamp_unix_unit(example).parser()
            except (ValueError, TypeError):
                example_ts = TimestampFormat.ISO.parser(self.timestamp_timezone)(example)
                return TimestampFormat.ISO.parser(self.timestamp_timezone or example_ts.tzinfo)

        return TimestampFormat.AUTO.parser(self.timestamp_timezone)

    def parse_timestamp(self, value: Any) -> Optional[datetime]:
        """Parse a timestamp, fall back to ISO."""
        if not value:
            return None
        if isinstance(value, datetime):
            return value
        try:
            return self.get_timestamp_parser()(value)
        except ValueError:
            return TimestampFormat.ISO.parser(self.timestamp_timezone)(value)

    def get_timestamp_from(self) -> Optional[datetime]:
        """Get the from timestamp, if any."""
        return self.parse_timestamp(self.timestamp_from)

    def get_timestamp_until(self) -> Optional[datetime]:
        """Get the until timestamp, if any."""
        return self.parse_timestamp(self.timestamp_until)


def _remove_prefix(metric_key: Optional[str]) -> Optional[str]:
    if metric_key and metric_key.startswith(METRIC_PREFIX):
        return metric_key[METRIC_PREFIX_LENGTH:]
    return metric_key


@dataclass
class ETLFile:
    """Defines a local ETL export/import file and workspace.

    Attributes:
        directory:
            The (local) directory used to store the ETL file and associated temporary files.
        prefix:
            The prefix used to create the ETL file. If not specified
            is defaulted to `import-{current timestamp}`.
    """

    directory: Optional[PathLike] = None
    prefix: Optional[str] = None

    def __post_init__(self):
        """Compute default prefix if not given as constructor parameter."""
        if not self.prefix:
            self.prefix = f'import-{datetime.utcnow():%Y%m%d.%H%M%S}'
        if not self.directory:
            self.directory = tempfile.mkdtemp(prefix='etl-import')

    @property
    def name(self) -> str:
        """Get the name of the ETL file."""
        return f'{self.prefix}-timeseries.csv.gz'

    @property
    def path(self) -> Path:
        """Get the full path of the ETL file."""
        assert self.directory is not None  # see __post_init__
        return Path(self.directory) / self.name


SeriesInput = Union[
    ETLFile,
    pd.DataFrame,
    str,
    os.PathLike,
    io.TextIOBase,
    Iterable[str],
    Iterable[Sequence[str]],
]


@dataclass
class WaylayETLSeriesImport:
    """A (reference to) a local file in the waylay ETL timeseries format and associated metadata."""

    series_input: Sequence[SeriesInput] = field(repr=False)
    import_file: ETLFile
    settings: SeriesSettings
    storage_bucket: str = ETL_IMPORT_BUCKET

    @property
    def storage_object_name(self):
        """Get default upload storage location."""
        return f'{ETL_IMPORT_UPLOAD_PREFIX}{self.import_file.name}'

    @property
    def name(self):
        """Get the name that identifies the import."""
        return self.import_file.prefix


@runtime_checkable
class CSVWriteable(Protocol):
    """Aything that can receive string writes."""

    def write(self, _: str) -> Any:
        """Write a string."""


CSVOutput = Union[
    str,
    os.PathLike,
    IO,
    CSVWriteable,
    ZipFile,
    TarFile
]


class ArchiveType(str, Enum):
    """Typed enumeration of supported archive import and export types."""

    description: str
    supports_multiple_files: bool
    expected_suffixes: List[str]
    excluded_suffixes: List[str]
    is_dir_type: bool

    def __new__(
        cls,
        value: str,
        description: str,
        supports_multiple_files: bool,
        expected_suffixes: Optional[List[str]] = None,
        excluded_suffixes: Optional[List[str]] = None,
        is_dir_type: bool = False,
    ):
        """Construct an enum value.

        See https://docs.python.org/3/library/enum.html#when-to-use-new-vs-init.
        """
        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        obj.supports_multiple_files = supports_multiple_files
        obj.expected_suffixes = expected_suffixes or []
        obj.excluded_suffixes = excluded_suffixes or []
        obj.is_dir_type = is_dir_type
        return obj

    TEXT = ('text', 'Single csv text file.', False, ['.csv'], ['.zip', '.tar', '.gz'])
    DIR = ('dir', 'Directory with csv text files.', True, None, None, True)
    DIR_GZ = ('dirgz', 'Directory with compressed csv text files.', True, None, None, True)
    ZIP = ('zip', 'Zip archive of csv text files.', True, ['.zip'])
    GZ = ('gz', 'Single compressed csv text file.', False, ['.gz'], ['.tar'])
    TAR = ('tar', 'Tar archive of csv text files.', True, ['.tar'], ['.gz'])
    TAR_GZ = ('targz', 'Compressed Tar archive of csv text files.', True, ['.tar', '.gz'])

    def __str__(self):
        """Get the string representation."""
        return self.value

    @classmethod
    def lookup(cls, value: Optional[str]) -> Optional['ArchiveType']:
        """Lookup an enum object from a str value."""
        if value is None:
            return None
        # workaround for lint and typing issues with replaced __new__ for enums
        # pylint: disable=no-value-for-parameter
        return ArchiveType(value.lower().replace('.', ''))  # type: ignore

    @classmethod
    def supported_for(cls, path: Path, multiple_files: bool) -> List['ArchiveType']:
        """Return a list of supported archive types, given a path and multiple_files flag."""
        if path.exists() and path.is_dir():
            return [
                at for at in list(ArchiveType)
                if at.is_dir_type
            ]
        archive_types_for_path = [
            at
            for at in list(ArchiveType)
            if at.expected_suffixes
            if all(s in path.suffixes for s in at.expected_suffixes)
            if not any(s in path.suffixes for s in at.excluded_suffixes)
        ]
        return [
            at for at in (archive_types_for_path or list(ArchiveType))
            if not multiple_files or at.supports_multiple_files
        ]

    @classmethod
    def for_import(cls, path: Path) -> Optional['ArchiveType']:
        """Return the archive type for import, as inferred from an existing file name."""
        if path.exists() and path.is_dir():
            return ArchiveType.DIR
        if path.suffixes and path.suffixes[-1] == '.tgz':
            return ArchiveType.TAR_GZ
        return next((
            at for at in list(ArchiveType)
            if at.expected_suffixes
            if all(s in path.suffixes for s in at.expected_suffixes)
            if not any(s in path.suffixes for s in at.excluded_suffixes)
        ), ArchiveType.TEXT)



"""Definitions for the 'query' entity of the 'queries' Service."""
from typing import Union, Dict

from .._base import WaylayResource
from .._decorators import (
    return_path_decorator
)
from ._decorators import (
    query_exception_decorator,
    query_return_dataframe_decorator,
    MultiFrameHandling,
)
from ._exceptions import (
    QueryRequestError, RestResponseError
)


CONFIG_ENTITY_DECORATORS = [
    query_exception_decorator,
    return_path_decorator(['query'])
]

CONFIG_LIST_DECORATORS = [
    query_exception_decorator,
    return_path_decorator(['queries', 'name'])
]

DATA_RESPONSE_DECORATORS = [
    query_exception_decorator,
    query_return_dataframe_decorator(
        'data',
        default_frames_handling=MultiFrameHandling.JOIN
    )
]

CONFIG_STATUS_DECORATORS = [
    query_exception_decorator,
    return_path_decorator([])
]

QUERY_NAME_ARG = {
    'name': 'name',
    'type': 'str',
    'description': 'Name for stored query.',
    'example': 'my_query_001'
}
QUERY_ENTITY_BODY = {
    'name': 'body',
    'type': 'dict',
    'description': 'A representation of a query entity.',
    'examples': ["""{
            "name": "max_flow_099",
            "query": {
                "freq": "PT1H",
                "aggregation": "max",
                "resource": "device_099",
                "data": [{ "metric": "flow"}]
            },
            "meta": { "description": "demo query" }
        }"""]
}
QUERY_DEFINITION_BODY = {
    'name': 'body',
    'type': 'dict',
    'description': 'A query definition.',
    'examples': ["""{
            "freq": "PT1H",
            "aggregation": "max",
            "resource": "device_099",
            "data": [{ "metric": "flow"}]
        }"""]
}
QUERY_ENTITY_OR_DEFINITION_BODY = {
    'name': 'body',
    'type': 'dict',
    'description': (
        'A representation of a query entity (including metadata), '
        'or the query definition itself.'
    ),
    'examples': [
        """{
            "name": "max_flow_099",
            "query": {
                "freq": "PT1H",
                "aggregation": "max",
                "resource": "device_099",
                "data": [{ "metric": "flow"}]
            },
            "meta": { "description": "demo query" }
        }""",
        """{
            "freq": "PT1H",
            "aggregation": "max",
            "resource": "device_099",
            "data": [{ "metric": "flow"}]
        }""",
    ]
}
QUERY_ENTITY_RESULT = {
    'name': 'query_definition',
    'type': 'Dict',
    'description': 'A representation of the stored query.',
}
ADDITIONAL_PARAMS_ARG = {
    'name': 'params',
    'type': 'dict',
    'description': (
        'Additional parameters, mapped to url query parameters. See API documentation.'
    ),
}


class QueryResource(WaylayResource):
    """REST Resource for the 'query' entity of the 'queries' Service."""

    link_roots = {
        'doc': '${doc_url}/api/query/',
        'apidoc':  '${apidoc_url}/queries.html'
    }

    actions = {
        'list': {
            'method': 'GET',
            'url': '/query',
            'arguments': [ADDITIONAL_PARAMS_ARG],
            'returns': [{
                'name': 'query_names',
                'type': 'List[str]',
                'description': 'A list of query names'
            }],
            'decorators': CONFIG_LIST_DECORATORS,
            'description': (
                'List the names of stored queries. '
                '<br>Use filter like <code>params=dict(q="name:demo")</code> to filter the listing. '
                '<br>Use <code>select_path=["queries"]</code> to return the query entities rather than names. '
            ),
            'links': {
                'doc': '?id=data-query-search-api',
                'apidoc': '#/query%20config/get_config_query'
            }
        },
        'create': {
            'method': 'POST',
            'url': '/query',
            'arguments': [QUERY_ENTITY_BODY],
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': (
                'Store a new query definition under a name. '
                'Fails if a query already exist with that name.'
            ),
            'links': {
                'doc': '?id=create',
                'apidoc': '#/query%20config/post_config_query'
            }
        },
        'get': {
            'method': 'GET',
            'url': '/query/{}',
            'arguments': [QUERY_NAME_ARG],
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Get the named query definition.',
            'links': {
                'doc': '?id=retrieve',
                'apidoc': '#/query%20config/get_config_query__query_name_'
            }
        },
        'remove': {
            'method': 'DELETE',
            'url': '/query/{}',
            'arguments': [QUERY_NAME_ARG],
            'decorators': CONFIG_STATUS_DECORATORS,
            'description': 'Remove the named query definition.',
            'links': {
                'doc': '?id=delete',
                'apidoc': '#/query%20config/delete_config_query__query_name_'
            }
        },
        'replace': {
            'method': 'PUT',
            'url': '/query/{}',
            'arguments': [QUERY_NAME_ARG, QUERY_ENTITY_OR_DEFINITION_BODY],
            'decorators': CONFIG_ENTITY_DECORATORS,
            'description': 'Replace the named query defition.',
            'links': {
                'doc': '?id=replace',
                'apidoc': '#/query%20config/put_config_query__query_name_'
            }
        },
        '_execute_by_name': {
            'method': 'GET',
            'url': '/data/{}',
            'arguments': [QUERY_NAME_ARG, ADDITIONAL_PARAMS_ARG],
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': (
                'Execute the timeseries query specified by the stored defintion of this name.'
            ),
            'links': {
                'doc': '?id=query-execution',
                'apidoc': '#/data/get_data_query__query_name_'
            }
        },
        '_execute_by_definition': {
            'method': 'POST',
            'url': '/data',
            'arguments': [QUERY_DEFINITION_BODY, ADDITIONAL_PARAMS_ARG],
            'decorators': DATA_RESPONSE_DECORATORS,
            'description': 'Execute the timeseries query specified in the request body.',
            'links': {
                'doc': '?id=query-execution',
                'apidoc': '#/data/post_data_query'
            }
        },
        'execute': {
            'arguments': [{
                'name': 'name_or_query',
                'type': 'Union[str, Dict]',
                'description': 'Either a name or query definition.'
            }, ADDITIONAL_PARAMS_ARG],
            'returns': [{
                'name': 'result',
                'type': 'pandas.DataFrame',
                'description': (
                    "A Pandas Dataframe containing the data, "
                    "unless 'response_constructor' specifies otherwise."
                )
            }],
            'wrapped_actions': ['_execute_by_name', '_execute_by_definition']
        },
        'create_or_replace': {
            'arguments': [
                {
                    'name': 'name',
                    'type': 'Optional[str]',
                    'description': 'Query name, if not given in body argument.'
                },
                QUERY_ENTITY_OR_DEFINITION_BODY,
                ADDITIONAL_PARAMS_ARG
            ],
            'wrapped_actions': ['create', 'replace']
        }
    }

    def execute(self, name_or_query: Union[str, Dict, None] = None, *, body: Dict = None, **kwargs):
        """Execute a timeseries query by name (string) or definition (object)."""
        if isinstance(name_or_query, str):
            return self._execute_by_name(name_or_query, **kwargs)  # pylint:disable=no-member

        # support query to be specified in the `body` argument
        query = name_or_query or body
        if isinstance(query, Dict):
            return self._execute_by_definition(body=query, **kwargs)  # pylint:disable=no-member

        raise QueryRequestError('The first argument should be a query name or definition.')

    def create_or_replace(
        self, name_or_query: Union[str, Dict, None] = None, *, body: Dict = None, **kwargs
    ):
        """Create or replace a query definition."""
        name = None
        if isinstance(name_or_query, str):
            name = name_or_query
        else:
            body = body or name_or_query
        if not body:
            raise QueryRequestError('Missing `body` argument with a query definition.')
        name = name or body.get('name')
        if not name:
            raise QueryRequestError(
                'Missing `name` as first argument or as property of the `body`.'
            )
        try:
            return self.replace(name, body=body, **kwargs)  # pylint:disable=no-member
        except RestResponseError as exc:
            if exc.response.status_code != 404:
                raise exc
            body = body if 'query' in body else dict(name=name, query=body)
            return self.create(body=body, **kwargs)  # pylint:disable=no-member



"""Base exception class hierarchy for errors in the waylay client."""
from simple_rest_client.exceptions import (
    ErrorWithResponse as _ErrorWithResponse,
    ClientConnectionError as _ClientConnectionError
)


class WaylayError(Exception):
    """Root class for all exceptions raised by this module."""


class AuthError(WaylayError):
    """Exception class for waylay authentication errors."""


class ConfigError(WaylayError):
    """Exception class for waylay client configuration."""


class RequestError(WaylayError):
    """Exception class for request validation errors within the waylay client.

    Notifies errors in tools and utilities that are not directly related to a REST call.
    """


class RestConnectionError(RequestError):
    """Exception class for request errors that are caused by client connection errors."""

    @classmethod
    def from_cause(cls, cause: _ClientConnectionError):
        """Convert a client connection error to a waylay client error."""
        return cls(cause)


class RestError(WaylayError):
    """Exception class for failures to make a REST call."""


class RestRequestError(RestError):
    """Exception class for failures to prepare a REST call."""


class RestResponseError(RestError):
    """Exception class wrapping the response data of a REST call."""

    def __init__(self, message, response):
        """Wrap a REST response in an error."""
        super().__init__(message)
        self.response = response

    @property
    def message(self):
        """Get the user message for this error."""
        return self.args[0]

    def __str__(self):
        """Render the error to a user-friendly string."""
        try:
            return (
                f"{self.__class__.__name__}({self.response.status_code}: " +
                f"'{self.message}'; {self.response.method} '{self.response.url}')"
            )
        except AttributeError:
            return format(self.response)

    def _get_from_body(self, key, default_value):
        # utility for subclasses in overriding the `message` method
        error_resp_body = self.response.body
        if isinstance(error_resp_body, dict):
            return error_resp_body.get(key, default_value)
        return default_value

    @classmethod
    def from_cause(cls, cause: _ErrorWithResponse):
        """Convert a rest client error to a waylay client error."""
        return cls(cause.message.split(',')[0], cause.response)


class RestResponseParseError(RestResponseError):
    """Exception raised when a successfull http request (200) could not be parsed succesfully."""



"""Client configuration."""

from typing import (
    Optional, Mapping, MutableMapping, Any, Type
)
import os
import re
from pathlib import Path
import json
import logging
import httpx
from appdirs import user_config_dir
from .auth import (
    DEFAULT_GATEWAY_URL,
    WaylayCredentials,
    NoCredentials,
    WaylayTokenAuth, WaylayToken,
    CredentialsCallback,
    parse_credentials,
)
from .auth_interactive import (
    ask_gateway,
    request_client_credentials_interactive,
    request_migrate_to_gateway_interactive,
    request_store_config_interactive,
    _root_url_for
)
from .exceptions import AuthError, ConfigError

log = logging.getLogger(__name__)

# http client dependencies
_http = httpx


def _http_get_global_settings(url, auth):
    return _http.get(url, auth=auth)


TenantSettings = Mapping[str, str]
Settings = MutableMapping[str, str]

DEFAULT_PROFILE = '_default_'
SERVICE_KEY_API = 'waylay_api'
SERVICE_KEY_GATEWAY = 'waylay_gateway'
SERVICE_KEY_ACCOUNTS = 'waylay_accounts'
DEFAULT_DOC_URL: str = 'https://docs.waylay.io/#'
DEFAULT_APIDOC_URL: str = 'https://docs.waylay.io/openapi/public/redocly'
DOC_URL_KEY: str = 'doc_url'
APIDOC_URL_KEY: str = 'apidoc_url'


class WaylayConfig():
    """Manages the authentication and endpoint configuration for the Waylay Platform."""

    profile: str
    _auth: WaylayTokenAuth
    _local_settings: Settings
    _tenant_settings: Optional[TenantSettings] = None

    _token_auth_provider: Type[WaylayTokenAuth] = WaylayTokenAuth

    def __init__(
        self, credentials: WaylayCredentials = None, profile: str = DEFAULT_PROFILE,
        settings: TenantSettings = None, fetch_tenant_settings=True,
        credentials_callback: Optional[CredentialsCallback] = None
    ):
        """Create a WaylayConfig."""
        self.profile = profile
        if credentials is None:
            credentials = NoCredentials()
        self._local_settings = {}
        if settings:
            self.set_local_settings(**settings)
        self._auth = self._token_auth_provider(
            credentials, credentials_callback=credentials_callback
        )
        if not fetch_tenant_settings:
            self._tenant_settings = {}

    @property
    def credentials(self):
        """Get current credentials.

        As configured or returned by last callback).
        """
        return self._auth.credentials

    def get_root_url(
        self, config_key: str, *,
        default_root_url: Optional[str] = None,
        gateway_root_path: Optional[str] = None,
        default_root_path: str = '',
        resolve_settings: bool = True
    ) -> Optional[str]:
        """Get the root url for a waylay service."""
        config_key = _root_url_key_for(config_key)
        # only resolve remote settings if no gateway available
        use_gateway = gateway_root_path is not None and self.gateway_url is not None
        default_url = (
            f'{self.gateway_url}{gateway_root_path}'
            if use_gateway
            else default_root_url
        )
        settings = self.local_settings
        if not use_gateway:
            resolve_settings = (
                resolve_settings
                # do not lookup settings for the bootstrap services
                and config_key not in (SERVICE_KEY_ACCOUNTS, SERVICE_KEY_API)
            )
            settings = self.get_settings(resolve=resolve_settings)
        url_override = settings.get(config_key)
        if url_override is not None:
            url_override = _root_url_for(url_override)
            if url_override.endswith(default_root_path):
                return url_override
            else:
                return (f'{url_override}{default_root_path}')
        if default_url is not None:
            return _root_url_for(default_url)
        if config_key == SERVICE_KEY_API:
            domain = self.get_valid_token().domain
            if domain:
                return f'{_root_url_for(domain)}/api'
        return None

    def set_root_url(self, config_key: str, root_url: Optional[str]):
        """Override the root url for the given server.

        Will persist on `save`.
        Setting a `None` value will remove the override.
        """
        config_key = _root_url_key_for(config_key)
        self.set_local_settings(**{config_key: root_url})

    @property
    def accounts_url(self) -> Optional[str]:
        """Get the accounts url."""
        url = self.credentials.accounts_url
        return _root_url_for(url) if url else None

    @property
    def gateway_url(self):
        """Get the gateway url."""
        url = self.credentials.gateway_url
        return _root_url_for(url) if url else None

    @property
    def doc_url(self) -> str:
        """Get the root url of the documentation site."""
        return self.settings.get(DOC_URL_KEY, DEFAULT_DOC_URL)

    @property
    def apidoc_url(self) -> str:
        """Get the root url of the api documentation site."""
        return self.settings.get(APIDOC_URL_KEY, DEFAULT_APIDOC_URL)

    @property
    def tenant_settings(self) -> TenantSettings:
        """Get the tenant settings as stored on accounts.

        Will fetch settings when not present and initialised with 'fetch_tenant_settings=True'.
        """
        if self._tenant_settings is None:
            self._tenant_settings = self._request_settings()

        return self._tenant_settings

    @property
    def local_settings(self) -> TenantSettings:
        """Get the settings overrides for this configuration.

        These include the endpoint overrides that are stored with the profile.
        """
        return self._local_settings

    def set_local_settings(self, **settings: Optional[str]) -> TenantSettings:
        """Set a local endpoint url override for a service."""
        for config_key, value in settings.items():
            if value is None and config_key in self._local_settings:
                del self._local_settings[config_key]
            if value is not None:
                self._local_settings[config_key] = value
        return self.local_settings

    @property
    def settings(self) -> TenantSettings:
        """Get settings, from tenant configuration and local overrides."""
        return self.get_settings(resolve=True)

    def get_settings(self, resolve=True) -> TenantSettings:
        """Get the tenant settings.

        As resolved form the accounts backend, and overridden with local settings.
        If `resolve=True`, fetch and cache tenant settings from the accounts backend.
        """
        return {
            **(self.tenant_settings if resolve else self._tenant_settings or {}),
            **self.local_settings
        }

    def get_valid_token(self) -> WaylayToken:
        """Get the current valid authentication token or fail."""
        if isinstance(self.auth, WaylayTokenAuth):
            try:
                return self.auth.assure_valid_token()
            except AuthError as exc:
                raise ConfigError(f'Cannot get valid token: {exc}') from exc
        raise ConfigError('not using token authentication')  # pragma: no cover

    @property
    def auth(self) -> _http.Auth:
        """Get the current the http authentication interceptor."""
        return self._auth

    @property
    def domain(self) -> Optional[str]:
        """Get the Waylay domain of the current user."""
        try:
            return self.get_valid_token().domain
        except ConfigError:
            return None

    @property
    def global_settings_url(self) -> str:
        """Get the REST url that fetches global settings."""
        root_url = self.get_root_url(
            'api', default_root_url=None, gateway_root_path='/configs/v1', resolve_settings=False
        )
        return f"{root_url}/settings"

    def _request_settings(self) -> TenantSettings:
        try:
            settings_resp = _http_get_global_settings(self.global_settings_url, auth=self.auth)
            settings_resp.raise_for_status()
            return {
                key: value
                for key, value in settings_resp.json().items()
                if key.startswith('waylay_')
            }
        except _http.HTTPStatusError as exc:
            if exc.response.status_code == 403:
                log.warning(
                    "You are not authorised to fetch tenant settings.\n"
                    "The Waylay SAAS defaults will be used, unless you\n"
                    "provide explicit overrides in the SDK Configuration profile."
                )
                return {}
            raise ConfigError('Cannot resolve tenant settings') from exc  # pragma: no cover
        except (_http.HTTPError, AuthError) as exc:
            raise ConfigError('Cannot resolve tenant settings') from exc  # pragma: no cover

    # config persistency methods
    @classmethod
    def config_file_path(cls, profile: str = DEFAULT_PROFILE) -> str:
        """Compute the default OS path used to store this configuration."""
        return os.path.join(user_config_dir('Waylay'), 'python_sdk', f".profile.{profile}.json")

    @classmethod
    def load(
        cls,
        profile: str = DEFAULT_PROFILE,
        *,
        interactive: bool = True,
        skip_error: bool = False,
        gateway_url: str = DEFAULT_GATEWAY_URL
    ):
        """Load a stored waylay configuration."""
        profile = DEFAULT_PROFILE if profile is None else profile
        try:
            with open(cls.config_file_path(profile), mode='r', encoding='utf-8') as config_file:
                config_json = json.load(config_file)
            waylay_config = cls.from_dict(config_json)
            if not waylay_config.gateway_url:
                msg = f"WaylayConfig profile '{profile}' uses a legacy accounts endpoint {waylay_config.accounts_url}."
                if interactive:
                    if request_migrate_to_gateway_interactive(profile, msg):
                        gateway_url = ask_gateway(waylay_config.accounts_url)
                        waylay_config.credentials.gateway_url = gateway_url
                        waylay_config.save()
                else:
                    log.warning(msg)
            return waylay_config
        except (FileNotFoundError, ValueError) as exc:
            msg = f"Config profile '{profile}' not found or invalid."
            if skip_error:
                log.warning(msg)
                return None
            if not interactive:
                raise ConfigError(msg) from exc
            credentials = request_client_credentials_interactive(default_gateway_url=gateway_url)
            instance = cls(
                credentials,
                profile=profile
            )
            request_store_config_interactive(profile, save_callback=instance.save)
            return instance

    @classmethod
    def from_dict(cls, config_json: Mapping[str, Any]):
        """Create a WaylayConfig from a dict representation."""
        config_json = dict(config_json)
        if 'credentials' in config_json:
            config_json['credentials'] = parse_credentials(config_json['credentials'])
        return cls(**config_json)

    def to_dict(self, obfuscate=True):
        """Get the (obfuscated) attributes of this WaylayConfig.

        Secret credentials are obfuscated.
        """
        return {
            'credentials': self.credentials.to_dict(obfuscate),
            'profile': self.profile,
            'settings': self.local_settings
        }

    def save(self) -> str:
        """Save the configuration as specified in the profile.

        Returns the save location.
        """
        config_path = Path(self.config_file_path(self.profile))
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, mode='w') as config_file:
            json.dump(self.to_dict(obfuscate=False), config_file)
            log.info("wrote waylay configuration: %s", config_path)
        return str(config_path)

    @classmethod
    def delete(cls, profile: str = DEFAULT_PROFILE) -> str:
        """Delete a stored profile.

        Returns the deleted location.
        """
        config_path = Path(cls.config_file_path(profile))
        if config_path.exists():
            config_path.unlink()
            log.warning("waylay configuration removed: %s", config_path)
        else:
            log.warning("waylay configuration not found: %s", config_path)
        return str(config_path)

    @classmethod
    def list_profiles(cls):
        """List stored config profiles."""
        config_dir = Path(cls.config_file_path()).parent
        return {
            profile_match[1]: str(config_file)
            for config_file in config_dir.iterdir()
            for profile_match in [re.match(r'\.profile\.(.*)\.json', config_file.name)]
            if profile_match
        }

    def __repr__(self):
        """Show the implementation class an main attributes."""
        return f'<WaylayConfig({str(self)})>'

    def __str__(self):
        """Show the main (obfuscated) attributes as a string."""
        return json.dumps(self.to_dict(obfuscate=True))


def _root_url_key_for(config_key: str):
    if config_key.startswith('waylay_'):
        return config_key
    return f"waylay_{config_key}"



"""Resource action method decorators that are generally usefull."""

import time
from functools import wraps
from typing import (
    Union, List, Mapping, Any, Callable, TypeVar, Dict, Optional
)

from logging import Logger, getLevelName, INFO
from simple_rest_client.exceptions import ErrorWithResponse, ClientConnectionError
from ..exceptions import RestResponseError, RestResponseParseError, RestConnectionError


def suppress_header_decorator(header_key):
    """Create a decorator that suppresses a configured header on a resource during execution."""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(slf, *args, **kwargs):
            header_value = slf.headers.pop(header_key, None)
            try:
                return action_method(slf, *args, **kwargs)
            finally:
                if header_value:
                    slf.headers[header_key] = header_value
        return wrapped
    return decorator


def exception_decorator(action_method):
    """Create a decorator that parses json error responses."""
    @wraps(action_method)
    def wrapped(*args, **kwargs):
        try:
            return action_method(*args, **kwargs)
        except ErrorWithResponse as exc:
            raise RestResponseError.from_cause(exc) from exc
        except ClientConnectionError as exc:
            raise RestConnectionError.from_cause(exc) from exc
    return wrapped


def default_params_decorator(default_params: Dict[str, Any]):
    """Create a decorator that initializes default url request parameters."""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            params = kwargs.pop('params', {})
            for key, value in default_params.items():
                if key not in params:
                    params[key] = value
            kwargs['params'] = params
            return action_method(*args, **kwargs)
        wrapped.__doc__ = (
            (wrapped.__doc__ or '') +
            "\nUses default url parameters (`params` argument):" +
            "".join(
                f"\n    '{key}': {default_value}"
                for key, default_value in default_params.items()
            ) + "\n"
        )
        return wrapped
    return decorator


def default_header_decorator(default_headers: Dict[str, Any]):
    """Create a decorator that initializes default header parameters."""
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            headers = kwargs.pop('headers', {})
            for key, value in default_headers.items():
                if key not in headers:
                    headers[key] = str(value)
            kwargs['headers'] = headers
            return action_method(*args, **kwargs)
        wrapped.__doc__ = (
            (wrapped.__doc__ or '') +
            "\nUses default headers (`headers` argument):" +
            "".join(
                f"\n    '{key}': {default_value}"
                for key, default_value in default_headers.items()
            ) + "\n"
        )
        return wrapped
    return decorator


def default_timeout_decorator(default_timeout: float):
    """Create a decorator that modifies the default http client timeout."""

    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            timeout = kwargs.pop('timeout', default_timeout)
            return action_method(*args, timeout=timeout, **kwargs)
        timeout_arg = f'timeout={default_timeout}'
        wrapped.__doc__ = (
            (wrapped.__doc__ or '') + f"""
Uses a modified default http client timeout:"
    {timeout_arg:30}: Client timeout for the REST request.
""")
        return wrapped
    return decorator


T = TypeVar('T')


def identity_transform(obj: T) -> T:
    """Return the argument."""
    return obj


def return_path_decorator(
    default_path: List[Union[int, str]],
    default_response_constructor: Optional[Callable[[Any], Any]] = None,
    response_constructor_advice: str = ''
):
    """Create a decorator method that extracts a part of the json body of an action response.

    The 'default_path' is a list of 'int' or 'str' keys that select through
    'dict' and 'list' structures.
    The option 'response_constructor' wrap the value in a class or constructor method
    It processes the following qualified arguments in a method call:
    * if 'raw=True' is set, the original response is returned
    * if 'select_path' is set, that path is used rather than the 'default_path'
       provided in the decorator constructor.
    * if 'response_constructor' is set, it is used to wrap the response
      rather than 'default_response_constructor'
    """
    def decorator(action_method):
        @wraps(action_method)
        def wrapped(*args, **kwargs):
            raw = kwargs.pop('raw', False)
            select_path = kwargs.pop('select_path', default_path)
            constructor = kwargs.pop(
                'response_constructor', default_response_constructor
            ) or identity_transform
            response = action_method(*args, **kwargs)
            if raw:
                return response
            if response.status_code == 204:
                # No-Content
                return None
            try:
                return constructor(
                    _select_path_from(response.body, [], select_path)
                )
            except AttributeError as exc:
                raise RestResponseParseError(exc.args[0], response) from exc
        resp_constr_arg = 'response_constructor'
        if default_response_constructor:
            resp_constr_arg += f"={_fqn_name(default_response_constructor)}"
        select_path_arg = f"select_path={default_path}"
        raw_arg = 'raw=False'
        dec_doc = (
            f"""
Returns the (parsed json) response body. Arguments that configure response handling:
    {select_path_arg:30}: extract path from the response body.
    {resp_constr_arg:30}: transforms the extracted response. {response_constructor_advice}
    {raw_arg:30}: If `True`, returns unmodified http response information.
"""
        )
        wrapped.__doc__ = (wrapped.__doc__ or '') + dec_doc
        return wrapped
    return decorator


def _fqn_name(func: Any) -> str:
    if func.__module__ == 'builtins':
        return func.__qualname__
    return f"{func.__module__}.{func.__qualname__}"


def return_body_decorator(action_method):
    """Create a decorator that returns the body of an action response.

    Skipped if 'raw=True' is provided in the call.
    """
    return return_path_decorator([])(action_method)


def _select_path_from(
    value: Any, ctx_path: List[Union[int, str]], path: List[Union[int, str]]
) -> Any:
    # Recursively select values from 'Mapping' (dict, 'str' key) input,
    # or positional elements from 'List' (list, 'int' key) input,
    # with a 'path' consisting of a list of keys.
    # Applied keys are moved to the 'ctx_path' to be used for error reporting.
    # 'str' keys applied to a 'List' are mapped through to the underlying structure.

    if not path:
        return value
    key = path[0]
    new_ctx_path = ctx_path + [key]
    rem_path = path[1:]
    if isinstance(value, List):
        if isinstance(key, int) and len(value) < key:
            return _select_path_from(value[key], new_ctx_path, rem_path)
        if isinstance(key, str):
            return [
                _select_path_from(element[key], new_ctx_path, rem_path)
                for element in value
            ]
    if isinstance(value, Mapping) and isinstance(key, str) and key in value:
        return _select_path_from(value[key], new_ctx_path, rem_path)

    ctx_path_str = '.'.join(str(p) for p in new_ctx_path)
    raise AttributeError(f"Cannot find '{ctx_path_str}' in response body.")


def log_server_timing_decorator(logger: Optional[Logger] = None, level: int = INFO):
    """Create a decorator that inspects and logs a 'Server-Timing' response header.

    This decorator should be below any response or exception decorator.
    """
    def decorator(action_method):
        logger.debug(
            'enabled Server-Timing logging for %s at %s',
            action_method.action, getLevelName(level)
        )

        def _log_server_timing(server_timing, start):
            if server_timing is None:
                end = time.time()
                logger.log(level, 'no-server-timing-available;dur=%.3f', (end-start) / 1000)
                return
            for line in server_timing.split(','):
                logger.log(level, line.strip())

        @wraps(action_method)
        def wrapped(*args, **kwargs):
            start = time.time()
            try:
                response = action_method(*args, **kwargs)
                _log_server_timing(response.headers.get('Server-Timing', None), start)
                return response
            except ErrorWithResponse as exc:
                _log_server_timing(exc.response.headers.get('Server-Timing', None), start)
                raise exc
            except Exception:
                _log_server_timing(None, start)
                raise
        wrapped.__doc__ = (
            (wrapped.__doc__ or '') +
            f"\nLogs server timing on {logger.name} at level {getLevelName(level)}\n"
        )
        return wrapped
    return decorator


def add_decorator_docs(decorators):
    """Add the documentation of the given decorators to a method.

    Intended for methods that delegate to action methods with the given decorators.
    """
    def update_doc(method):
        method.__doc__ = (method.__doc__ or '') + "\n\n".join(
            _extract_doc(dec) for dec in decorators
        )
        return method

    return update_doc


def _extract_doc(decorator):
    def _dummy(): pass
    return decorator(_dummy).__doc__ or ''



"""Utilities."""

from io import TextIOWrapper
from tempfile import SpooledTemporaryFile


class NonClosingTextIOWrapper(TextIOWrapper):
    """TextIOWrapper that doesn't close, but detaches the wrapped buffer when gc-ed."""

    # SEE https://bugs.python.org/issue21363
    # https://github.com/astropy/astropy/pull/11809/files#r644158123
    def __del__(self):
        """Detach from underlying buffer on deletion."""
        self.detach()


class WrappeableSpooledTemporaryFile(SpooledTemporaryFile):
    """Adapted SpooledTemporaryFile so it can be wrapped in a TextIOWrapper."""

    def readable(self) -> bool:
        """Assert readability."""
        return True

    def writable(self) -> bool:
        """Assert writability."""
        return True

    def seekable(self) -> bool:
        """Assert seekablity."""
        return True



"""Module that implements export readers for waylay timeseries data."""
from datetime import datetime
from dataclasses import replace
from io import TextIOBase, IOBase
from os import PathLike
from tarfile import TarFile
from time import time
from typing import (
    Any, Callable, Dict, Iterator, Iterable, List,
    Optional, Sequence, TextIO, Tuple, Union, IO,
    cast, BinaryIO
)
from zipfile import ZipFile

import collections.abc
import pathlib
import gzip
import csv
import urllib.parse
import logging
import os

import pandas as pd

from waylay.service.data.series import SeriesResource

from .model import (
    ArchiveType, CSVWriteable, Measurement, MeasurementIterator,
    SeriesIterator, SeriesProvider, SeriesSettings, CSVOutput, TimestampFormat
)
from .util import (
    NonClosingTextIOWrapper,
    WrappeableSpooledTemporaryFile,
)
LOG = logging.getLogger(__name__)
DEFAULT_EXPORT_PAGE_SIZE = 2000
DEFAULT_SPOOL_MAX_SIZE = 2 << 18
TS_FORMAT_MILLIS = TimestampFormat.MILLIS.formatter()


def create_export_series_provider(
    settings: SeriesSettings, series_api: SeriesResource
) -> SeriesProvider:
    """Create a series provider from the waylay api for the given settings."""
    return PagingExportSeriesProvider(settings, series_api)


class PagingExportSeriesProvider(collections.abc.Mapping):
    """A series provider that uses paging exports from the waylay series api."""

    def __init__(self, settings: SeriesSettings, series_api: SeriesResource):
        """Create an export series provider."""
        self.settings = settings
        self.series_api = series_api
        self._keys = [
            (resource.id, metric.name)
            for resource in settings.iter_resources()
            for metric in settings.iter_metrics()
        ]

    def __iter__(self):
        """Iterate over the (resource, metric) keys of this provider."""
        return iter(self._keys)

    def __len__(self):
        """Get the number of requested (resource,metric) combinations."""
        return len(self._keys)

    def __getitem__(self, resource_metric: Tuple[str, str]) -> Iterable[Measurement]:
        """Create an measurement iterable for a given resource, metric."""
        resource, metric = resource_metric
        provider = self

        class _MeasurementIterable():
            def __iter__(self) -> Iterator[Measurement]:
                return provider._iter_single_series_export(resource, metric)

        return _MeasurementIterable()

    def _iter_single_series_export(
        self,
        resource_id: str, metric: str,
        page_size: int = DEFAULT_EXPORT_PAGE_SIZE
    ) -> MeasurementIterator:
        """Paging iterator over the specified resource and metric."""
        params: Dict[str, Any] = {
            'from': 0,
            'order': 'ascending',
            'limit': page_size
        }
        settings = self.settings
        timestamp_from = settings.get_timestamp_from()
        if timestamp_from:
            params['from'] = TS_FORMAT_MILLIS(timestamp_from)
        timestamp_until = settings.get_timestamp_until()
        if timestamp_until:
            params['until'] = TS_FORMAT_MILLIS(timestamp_until)

        while True:
            LOG.debug(f'fetching a page of series for %s:%s', resource_id, metric)
            resp = self.series_api.export(resource_id, metric, params=params, select_path=None)
            if 'series' in resp:
                for item in resp['series']:
                    yield (
                        pd.Timestamp(item[0], unit='ms', tz='UTC'),
                        item[1]
                    )
            next_link = resp.get('_links', {}).get('next', {}).get('href')
            if not next_link:
                return
            next_query = urllib.parse.urlparse(next_link).query
            params = urllib.parse.parse_qs(next_query)


def _iter_settings(settings: SeriesSettings):
    if settings.per_resource:
        for resource in settings.iter_resources():
            name = f'{settings.name}-{resource.id}' if settings.name else resource.id
            entry_settings = replace(
                settings, name=name, resource=resource.id, per_resource=False
            )
            yield from _iter_settings(entry_settings)
    elif settings.per_metric:
        for metric in settings.iter_metrics():
            name = f'{settings.name}-{metric.name}' if settings.name else metric.name
            entry_settings = replace(
                settings, name=name, metric=metric.name, per_metric=False
            )
            yield from _iter_settings(entry_settings)
    else:
        yield settings.name or 'series', settings


ARCHIVE_DEFAULT_EXTENSIONS = {
    '.zip': 'zip'
}


def export_csv(
    output: CSVOutput,
    settings: SeriesSettings,
    series: SeriesProvider,
    archive_type: Optional[ArchiveType] = None,
    progress: bool = True,
):
    """Export a series iterator to a CSV file or archive."""
    if isinstance(output, (str, PathLike)):
        mkdirs = os.fspath(output).endswith('/')
        output = pathlib.Path(output)
        if mkdirs:
            output.mkdir(parents=True, exist_ok=True)
        supported_archive_types = ArchiveType.supported_for(output, settings.per_resource or settings.per_metric)
    elif isinstance(output, ZipFile):
        supported_archive_types = [ArchiveType.ZIP]
    elif isinstance(output, TarFile):
        supported_archive_types = [ArchiveType.TAR, ArchiveType.TAR_GZ]
    elif isinstance(output, TextIOBase):
        supported_archive_types = [ArchiveType.TEXT]
    elif isinstance(output, IOBase):
        supported_archive_types = [
            at for at in ArchiveType
            if at.supports_multiple_files or not (settings.per_resource or settings.per_metric)
            if not at.is_dir_type
        ]
    elif isinstance(output, CSVWriteable):
        supported_archive_types = [ArchiveType.TEXT]
    else:
        supported_archive_types = []

    if not supported_archive_types:
        raise TypeError(f'Unsupported output type {type(output)}')

    archive_type = archive_type or supported_archive_types[0]
    LOG.debug('output = %r', output)
    LOG.debug('archive_type = %r', archive_type)

    if (settings.per_resource or settings.per_metric) and not archive_type.supports_multiple_files:
        raise ValueError(f'Archive type `{archive_type}` does not support multiple outputs.')

    if archive_type not in supported_archive_types:
        raise ValueError(f'Archive type `{archive_type}` not supported for given output')

    if archive_type in (ArchiveType.TEXT, ArchiveType.GZ):
        LOG.info('exporting to single output: %s', output)
        if isinstance(output, TextIOBase):
            return export_csv_to_file(output, settings, series)
        if isinstance(output, IOBase):
            if archive_type is ArchiveType.GZ:
                with gzip.open(output, mode='wt') as text_output:
                    return export_csv_to_file(cast(TextIO, text_output), settings, series)
            return export_csv_to_file(NonClosingTextIOWrapper(cast(IO[bytes], output)), settings, series)
        if isinstance(output, CSVWriteable):
            return export_csv_to_file(output, settings, series)
        if isinstance(output, pathlib.Path):
            if archive_type is ArchiveType.GZ:
                with gzip.open(output, mode='wt') as text_output:
                    return export_csv_to_file(text_output, settings, series)
            with open(output, 'wt') as text_output:
                return export_csv_to_file(text_output, settings, series)
        raise TypeError(f'Unsupported output type for archive type {archive_type}')  # pragma: no cover

    if archive_type in (ArchiveType.DIR, ArchiveType.DIR_GZ):
        LOG.info('exporting to directory: %s', output)
        if isinstance(output, pathlib.Path):
            return export_csv_to_dir(
                output, settings, series,
                compress=archive_type == ArchiveType.DIR_GZ
            )
        raise TypeError(f'Unsupported output type for archive type {archive_type}')  # pragma: no cover

    if archive_type is ArchiveType.ZIP:
        LOG.info('exporting to zip archive: %s', output)
        if isinstance(output, ZipFile):
            return export_csv_to_zip_archive(output, settings, series)
        if isinstance(output, pathlib.Path):
            with ZipFile(output, mode='w') as zip_file:
                return export_csv_to_zip_archive(zip_file, settings, series)
        if isinstance(output, IOBase) and not isinstance(output, TextIOBase):
            zip_file = ZipFile(cast(IO[bytes], output), mode='w')
            return export_csv_to_zip_archive(zip_file, settings, series)
        raise TypeError(f'Unsupported output type for archive type {archive_type}')  # pragma: no cover

    if archive_type in (ArchiveType.TAR, ArchiveType.TAR_GZ):
        LOG.info('exporting to tar archive: %s', output)
        if isinstance(output, TarFile):
            return export_csv_to_tar_archive(output, settings, series)
        tar_mode = 'w' if archive_type is ArchiveType.TAR else 'w:gz'
        if isinstance(output, pathlib.Path):
            with TarFile.open(output, mode=tar_mode) as tar_file:
                return export_csv_to_tar_archive(tar_file, settings, series)
        if isinstance(output, IOBase) and not isinstance(output, TextIOBase):
            tar_file = TarFile.open(fileobj=cast(IO[bytes], output), mode=tar_mode)
            return export_csv_to_tar_archive(tar_file, settings, series)
        raise TypeError(f'Unsupported output type for archive type {archive_type}')  # pragma: no cover


def export_csv_to_dir(
    csv_dir: Union[str, PathLike], settings: SeriesSettings,
    series: SeriesProvider, compress: bool = False
):
    """Export a series provider to a directory csv files."""
    dir_path = pathlib.Path(csv_dir)
    # raises a File exists if `dir` is a file
    dir_path.mkdir(parents=True, exist_ok=True)
    for entry_name, entry_settings in _iter_settings(settings):
        if compress:
            file_name = dir_path / f'{entry_name}.csv.gzip'
            LOG.info('exporting to compressed file: %s', file_name)
            with gzip.open(file_name, 'wt') as file:
                export_csv_to_file(file, entry_settings, series)
        else:
            file_name = dir_path / f'{entry_name}.csv'
            LOG.info('exporting to uncompressed file: %s', file_name)
            with open(dir_path / f'{entry_name}.csv', 'w') as file:
                export_csv_to_file(file, entry_settings, series)


def export_csv_to_zip_archive(
    zip_file: ZipFile, settings: SeriesSettings,
    series: SeriesProvider
):
    """Export a series provider to a zip archive of csv files."""
    for entry_name, entry_settings in _iter_settings(settings):
        entry_file_name = f'{entry_name}.csv'
        LOG.info('writing zip entry: %s', entry_file_name)
        with zip_file.open(entry_file_name, 'w') as zip_file_entry:
            export_csv_to_file(NonClosingTextIOWrapper(zip_file_entry), entry_settings, series)


def export_csv_to_tar_archive(
    tar_file: TarFile, settings: SeriesSettings, series: SeriesProvider
):
    """Export a series provider to a tar archive of csv files per resource."""
    for entry_name, entry_settings in _iter_settings(settings):
        entry_file_name = f'{entry_name}.csv'
        LOG.info('writing tar entry: %s', entry_file_name)
        with WrappeableSpooledTemporaryFile(mode='wb+', max_size=DEFAULT_SPOOL_MAX_SIZE) as csv_buffer:
            export_csv_to_file(NonClosingTextIOWrapper(csv_buffer), entry_settings, series)
            tar_info = TarFile.tarinfo(f'{entry_name}.csv')
            tar_info.mtime = int(time())
            tar_info.size = csv_buffer.tell()
            csv_buffer.seek(0)
            tar_file.addfile(tar_info, csv_buffer)


def export_csv_to_file(
    output: CSVWriteable, settings: SeriesSettings, series_provider: SeriesProvider
):
    """Export a series provider to a CSV output stream according to the settings."""
    if settings.value_column:
        _export_csv_values(output, settings, series_provider)
    else:
        _export_csv_columns(output, settings, series_provider)


def _export_csv_values(
    output: CSVWriteable, settings: SeriesSettings, series_provider: SeriesProvider
):
    selected_resource_ids = [r.id for r in settings.iter_resources()]
    selected_metric_names = [m.name for m in settings.iter_metrics()]

    header = []
    write_resource = False
    write_metric = False
    write_timestamp = False
    format_timestamp = settings.get_timestamp_formatter()
    if settings.resource_column or len(selected_resource_ids) > 1:
        write_resource = True
        header.append(settings.resource_column or 'resource')
    if settings.metric_column or len(selected_metric_names) > 1:
        write_metric = True
        header.append(settings.metric_column or 'metric')
    if settings.timestamp_column:
        write_timestamp = True
        header.append(settings.timestamp_column or 'timestamp')
    header.append(settings.value_column or 'value')
    csv_writer = csv.writer(output, delimiter=',', quoting=csv.QUOTE_MINIMAL)
    if settings.write_csv_header:
        csv_writer.writerow(header)
        LOG.debug('csv header = %r', header)
    for (resource_id, metric_name), measures in series_provider.items():
        if resource_id not in selected_resource_ids:
            continue
        if metric_name not in selected_metric_names:
            continue
        rec = []
        if write_resource:
            rec.append(settings.resource_for(resource_id).key_or_id)
        if write_metric:
            rec.append(settings.metric_for(metric_name).key_or_name)
        if write_timestamp:
            for timestamp, value in measures:
                csv_writer.writerow((*rec, format_timestamp(timestamp), str(value)))
        else:
            for timestamp, value in measures:
                csv_writer.writerow((*rec, str(value)))


def _export_csv_columns(output: CSVWriteable, settings: SeriesSettings, series: SeriesProvider):
    selected_resource_ids = [r.id for r in settings.iter_resources()]
    LOG.debug('selected resource ids = %r', selected_resource_ids)
    selected_metric_names = [m.name for m in settings.iter_metrics()]
    LOG.debug('selected metric names = %r', selected_metric_names)

    header = []
    write_timestamp = False
    write_resource = False
    format_timestamp = settings.get_timestamp_formatter()
    if settings.resource_column or len(selected_resource_ids) > 1:
        write_resource = True
        header.append(settings.resource_column or 'resource')
    if settings.timestamp_column:
        write_timestamp = True
        header.append(settings.timestamp_column or 'timestamp')
    header.extend([m.key_or_name for m in settings.iter_metrics()])

    csv_writer = csv.writer(output, delimiter=',', quoting=csv.QUOTE_MINIMAL)
    LOG.debug('csv header = %r', header)
    if settings.write_csv_header:
        csv_writer.writerow(header)
    for resource_id in selected_resource_ids:
        resource_key = settings.resource_for(resource_id).key_or_id
        row_iterator = _merge_iterators(
            *(
                iter(series.get((resource_id, metric_name), []))
                for metric_name in selected_metric_names
            ),
            format_timestamp=format_timestamp if write_timestamp else None
        )
        if write_resource:
            for row in row_iterator:
                csv_writer.writerow((resource_key, *row))
        else:
            for row in row_iterator:
                csv_writer.writerow(row)


def _merge_iterators(
    *measure_iterators: MeasurementIterator,
    format_timestamp: Optional[Callable[[datetime], Any]],
) -> Iterator[Sequence[Any]]:
    """Merge a sequence of MeasurementIterators to a sequence of csv output records.

    Each iterator is assumed to be time-ordered.
    Each different timestamp will yield a seperate row.
    """
    next_measurements = list(next(it, None) for it in measure_iterators)
    while True:
        try:
            current_timestamp = min(m[0] for m in next_measurements if m is not None)
        except ValueError:
            # no timestamps left
            return
        curr_record: List[Optional[Any]] = [format_timestamp(current_timestamp)] if format_timestamp else []
        for idx, measurement in enumerate(next_measurements):
            if measurement is None:
                curr_record.append(None)
                continue
            timestamp, value = measurement
            if timestamp == current_timestamp:
                curr_record.append(value)
                next_measurements[idx] = next(measure_iterators[idx], None)
            else:
                curr_record.append(None)
        yield curr_record



"""REST definitions for the 'subscription' entity of the 'storage' service."""
from urllib.parse import quote

from waylay.service import WaylayResource
from waylay.service import decorators


BUCKET_NAME_ARG = {
    'name': 'bucket_name',
    'type': 'str',
    'description': 'Name of a Waylay storage bucket.',
    'examples': ['assets', 'public']
}
SUBSCRIPTION_ID_ARG = {
    'name': 'subscription_id',
    'type': 'str',
    'description': 'The identifier of a subscription.',
    'examples': ['my_notification_subscription_349']
}
ADDITIONAL_FILTER_PARAMS_ARG = {
    'name': 'params',
    'type': 'dict',
    'description': (
        "Additional parameters, like 'prefix', 'suffix', 'event_type', "
        "'channel_type', 'channel_id', that may filter subscriptions. "
        "See API documentation."
    )
}
SUBSCRIPTION_ENTITY_BODY = {
    'name': 'body',
    'type': 'dict',
    'description': 'A JSON representation of a subscription. See API documentation.'
}


class SubscriptionResource(WaylayResource):
    """REST Resource for the 'subscription' entity of the 'storage' service."""

    link_roots = {
        'doc': '${doc_url}/api/storage/?id=',
        'apidoc':  '${apidoc_url}/storage.html',
    }
    actions = {
        'list': {
            'method': 'GET', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['subscriptions']),
            ],
            'arguments': [BUCKET_NAME_ARG, ADDITIONAL_FILTER_PARAMS_ARG],
            'description': 'List available subscriptions for a given bucket.',
            'links': {
                'doc': 'list-bucket-subscriptions',
                'apidoc': '',
            }
        },
        'get': {
            'method': 'GET', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG, SUBSCRIPTION_ID_ARG],
            'description': 'Retrieve the representation of a notification subscription.',
            'links': {
                'doc': 'get-subscription',
                'apidoc': '',
            }
        },
        'create': {
            'method': 'POST', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG, SUBSCRIPTION_ENTITY_BODY],
            'description': 'Create a new notification subscription.',
            'links': {
                'doc': 'create-subscription',
                'apidoc': '',
            }
        },
        'replace': {
            'method': 'PUT', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG, SUBSCRIPTION_ID_ARG, SUBSCRIPTION_ENTITY_BODY],
            'description': 'Create or Replace the definition of a notification subscription.',
            'links': {
                'doc': 'update-subscription',
                'apidoc': '',
            }
        },
        'remove': {
            'method': 'DELETE', 'url': '/subscription/{}/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG, SUBSCRIPTION_ID_ARG],
            'description': 'Remove a notification subscription.',
            'links': {
                'doc': 'delete-subscription',
                'apidoc': '',
            }

        },
        'remove_all': {
            'method': 'DELETE', 'url': '/subscription/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG, ADDITIONAL_FILTER_PARAMS_ARG],
            'description': 'Remove all notification subscription that satisfy a query.',
            'links': {
                'doc': 'delete-subscriptions',
                'apidoc': '',
            }
        },
    }



"""Internal service with tools for timeseries import and export."""

from waylay.service import WaylayService, WaylayServiceContext
from waylay.config import WaylayConfig

from .tool import TimeSeriesETLTool


class TimeSeriesService(WaylayService):
    """Tool service for the timeseries import and export operations."""

    service_key = 'timeseries'

    etl_tool: TimeSeriesETLTool

    resource_definitions = {
        'etl_tool': TimeSeriesETLTool
    }

    def configure(self, config: WaylayConfig, context: WaylayServiceContext) -> 'TimeSeriesService':
        """Configure endpoints and authentication with given config."""
        super().configure(config, context)
        self.etl_tool.set_context(context)
        return self



"""Command line interface for the series service."""
from typing import Any, Dict, List, Union, TextIO
import sys
import io
import pathlib
import logging

from argparse import ArgumentParser, Namespace, RawTextHelpFormatter
from waylay.service.timeseries.tool import ResourceUpdateLevel

from waylay import WaylayClient
from waylay.service.timeseries.parser.model import (
    SeriesSettings, ArchiveType, TimestampFormat, Metric, Resource
)

from ._decorators import cli_exeption_decorator

LOG = logging.getLogger(__name__)

SERIES_COMMANDS = [CMD_EXPORT, CMD_IMPORT] = ['export', 'import']


def init_series_parser(parser: ArgumentParser):
    """Initialize the parser for the 'series' command."""
    if sys.version_info > (3, 7):
        series_cmd_parsers = parser.add_subparsers(
            help='Get or add timeseries data',
            dest='series_cmd',
            title='Series command',
            required=True
        )
    else:
        series_cmd_parsers = parser.add_subparsers(dest='series_cmd', title='Series command')

    init_series_export_parser(series_cmd_parsers.add_parser(
        CMD_EXPORT,
        description='Export series to a local file',
        help='Export series to a local file',
        formatter_class=RawTextHelpFormatter
    ))
    init_series_import_parser(series_cmd_parsers.add_parser(
        CMD_IMPORT,
        description='Import series from a local file',
        help='Import series from a local file',
        formatter_class=RawTextHelpFormatter
    ))
    return parser


def init_series_timestamp_parser(parser: ArgumentParser):
    """Initialize the parser for the timestamp filtering and mapping."""
    parser.add_argument(
        '-t', '--timestamp-format', dest='timestamp_format', nargs='?', default=TimestampFormat.ISO.value,
        type=TimestampFormat.lookup, choices=list(t.value for t in TimestampFormat),
        help=(
            'Specify a timestamp format for csv export rendering and csv import parsing.\n'
            + '\n'.join(f'- {t.value: <6}: {t.description}' for t in TimestampFormat)
        ),
    )
    parser.add_argument(
        '-z', '--timezone', dest='timestamp_timezone', metavar='TZ', nargs='?', default='UTC',
        help=(
            'Specify the timezone used for rendering iso-formatted timestamps,'
            "\nand for parsing inputs that do not specify a timezone."
        ),
    )
    parser.add_argument(
        '--from', dest='timestamp_from', metavar='FROM', nargs='?', default=None,
        help=(
            'Filter the series to have a timestamp equal to or greater than this one.'
        ),
    )
    parser.add_argument(
        '--until', dest='timestamp_until', metavar='UNTIL', nargs='?', default=None,
        help=(
            'Filter the series to have a timestamp strictly less than this one.'
        ),
    )


def init_series_export_parser(parser: ArgumentParser):
    """Initialize the parser for the 'series export' command."""
    parser.add_argument(
        '-r', '--resources', dest='resources', metavar='RESOURCE_ID[=resource_key]', nargs='+',
        help=(
            'Specify the resource ids to export.'
            "\nA resource id can be mapped with a 'RESOURCE_ID[=resource_key]' expression, with"
            "\n    'RESOURCE_ID' the resource id in waylay time series database"
            "\n    'resource_key' the value written to the csv file"
            '\nWhen specified, time series with resource ids not mentioned will be skipped.'
        ),
        required=True,
    )
    parser.add_argument(
        '-m', '--metrics', dest='metrics', metavar='METRIC_NAME[=metric_key]', nargs='+',
        help=(
            'Specify the metric names to export'
            "\nA metric name can be mapped with a 'METRIC_NAME[=metric_key]' expression, with"
            "\n    'METRIC_NAME' the metric name used in waylay time series database"
            "\n    'metric_key' the value used in the exported csv file "
            '(as metric header or value in a metric column)'
            '\nWhen specified, time series with metric name/keys not mentioned will be skipped.'
        ),
        required=True,
    )
    parser.add_argument(
        '-c', '--columns', dest='columns', metavar='COL[=name]',  nargs='*',
        help=(
            'Specify the output columns for the export.'
            '\n - resource[=column_name]'
            "\n     Include a 'resource' column for the resource ids."
            '\n     By default only present when multiple resources are exported.'
            "\n     'resource' will always include the resource column."
            "\n     'resource=' will always omit the resource column."
            "\n     'resource=office' will rename the resource column to 'office'."
            '\n - value[=column_name]'
            "\n     Include/rename a 'value' column for the series values."
            '\n     Setting this implies an import/export in a normalized format,'
            '\n     with a single metric per row.'
            '\n - metric[=column_name]'
            "\n     Include/rename a 'metric' column for the metric name."
            "\n     Not used unless the 'value' column is set."
            '\n - timestamp[=column_name]'
            "\n     Rename a 'timestamp' column for the timestamps."
            "\n     'timestamp=' will omit the default timestamp column."
        ),
    )
    parser.add_argument(
        '-o', '--output', dest='output', nargs='?', default=sys.stdout,
        help=(
            'Specify an output for the export (file or dir).\n'
            'Defaults to standard output (csv text).'
        ),
    )
    parser.add_argument(
        '-a', '--archive-type', dest='archive_type', nargs='?', default=None,
        type=ArchiveType.lookup, choices=list(t.value for t in ArchiveType),  # type: ignore
        help=(
            'Specify an archive type for the export.\n'
            + '\n'.join(f'- {t.value: <6}: {t.description}' for t in ArchiveType)
            + '\nIf not specified, the archive type is inferred from the output name or type.'
        ),
    )
    init_series_timestamp_parser(parser)
    parser.add_argument(
        '--per-resource', dest='per_resource', action="store_true", default=False,
        help=(
            'If set, export each resource in a seperate file.\n'
            'Does not work for single-file archive types (text, gz).'
        ),
    )
    parser.add_argument(
        '--per-metric', dest='per_metric', action="store_true", default=False,
        help=(
            'If set, export each metric in a seperate file.\n'
            'Does not work for single-file archive types (text, gz).'
        ),
    )
    parser.add_argument(
        '--silent', dest='silent', action="store_true", default=False,
        help=(
            "Minimize stderr output (no progress bar)."
        ),
    )


def init_series_import_parser(parser: ArgumentParser):
    """Initialize the parser for the 'series import' command."""
    parser.add_argument(
        'series', nargs='+',
        help=(
            'Specify an input for the import (file or dir).\n'
            "Use '-' to specify standard input (csv text)."
        )
    )
    parser.add_argument(
        '-n', '--name', dest='name', nargs='?', default=None,
        help=(
            'ETL import job name, used as prefix base name for the ETL import file.'
        ),
    )
    parser.add_argument(
        '-R', '--default-resource', metavar='RESOURCE_ID', dest='default_resource', nargs='?', default=None,
        help=(
            'A default resource id for this import. Only used when not specified in the import file.'
            '\nWhen not set, and not provided by the import file, the default resource name'
            "\nis inferred from the base name of the import file or from the '--name' argument."
        ),
    )
    parser.add_argument(
        '-M', '--default-metric', metavar='METRIC_NAME', dest='default_metric', nargs='?', default=None,
        help=(
            'A default metric name for this import. Only used when not specified as import data.'
        )
    )
    parser.add_argument(
        '-r', '--resources', dest='resources', metavar='RESOURCE_ID[=resource_key]', nargs='*',
        help=(
            'Filter and/or rename the resource ids in the import.'
            "\nA resource id can be mapped with a 'RESOURCE_ID[=resource_key]' expression, with"
            "\n    'RESOURCE_ID' the resource id written to the waylay time series database"
            "\n    'resource_key' the value used in the import csv file"
            '\nWhen specified, time series with resource id/keys not mentioned will be skipped.'
        ),
    )
    parser.add_argument(
        '-m', '--metrics', dest='metrics', metavar='METRIC_NAME[=metric_key]', nargs='*',
        help=(
            'Specify the metric names to import'
            "\nA metric name can be mapped with a 'METRIC_NAME[=metric_key]' expression, with"
            "\n    'METRIC_NAME' the metric name used in waylay time series database"
            "\n    'metric_key' the value used in the import csv file "
            '(as metric header or value in a metric column)'
            '\nWhen specified, time series with metric name/keys not mentioned will be skipped.'
        ),
    )
    parser.add_argument(
        '-c', '--columns', dest='columns', metavar='COLUMN[=column_name]',  nargs='*',
        help=(
            'Specify alternate column names in the import files.'
            "\nBy default, import csv files are expected to have a 'timestamp' column, and"
            "\n all other columns contain series with metric name in the header."
            '\n - resource[=resource_column_name]'
            "\n     Use the 'resource_column_name' (default 'resource') as keys for resource ids."
            "\n     If not set, any column named 'resource' will be used, or"
            "\n       the default resource key (see '--default-resource') will be used."
            "\n     E.g. 'resource=device' will expect a column named 'device' containing resource keys'."
            '\n - value[=value_column_name]'
            "\n     Use the 'value_column_name' (default 'value') column for the series values."
            '\n     Setting this implies an a normalized format with one metric per row.'
            '\n - metric[=metric_column_name]'
            "\n     Use the 'metric_column_name' (default 'metric') column for the series metric keys."
            "\n     Only relevant when 'value' column is set."
            "\n     When not set, any column 'metric' is used, or a"
            '\n     default metric name (see --default-metric) should be specified.'
            '\n - timestamp[=timestamp_column_name]'
            "\n     Use the 'timestamp_column_name' (default 'timestamp') column as series timestamp."
        ),
    )
    init_series_timestamp_parser(parser)
    parser.add_argument(
        '-T', '--temp-dir', dest='temp_dir', nargs='?', default=None,
        help=(
            'The local storage location to use in preparing the ETL file.\n'
            "Defaults to a system generated temp file with prefix 'etl-import'."
        ),
    )
    parser.add_argument(
        '--resource-update', dest='resource_update_level', nargs='?',
        type=ResourceUpdateLevel, default=ResourceUpdateLevel.ID,
        choices=list(level.value for level in ResourceUpdateLevel),
        help=(
            'Level of update of the waylay resource metadata after import.'
            '\n- none : Do not update resource metadata'
            '\n- id   : (default) Make sure the resource exists (id only)'
            '\n- all  : Create or update id, name and metrics metadata on the resource.'
        )
    )
    parser.add_argument(
        '--silent', dest='silent', action="store_true", default=False,
        help=(
            "Minimize stderr output (no progress bar)"
        ),
    )
    parser.add_argument(
        '--no-import', dest='no_import', action="store_true", default=False,
        help=(
            'Process the inputs without uploading to the timeseries database.'
            '\nOutputs the location of the (temporary) etl import file.'
        ),
    )


def init_series_import_timestamp_generation(parser: ArgumentParser):
    """Initialize the parser argument for generation/mapping of timestamps during import."""
    parser.add_argument(
        '--first', dest='timestamp_first', nargs='?', default=None,
        help=(
            'Forces the first timestamp, and increments'
            'the following timestamps with the same amount.'
        )
    )
    parser.add_argument(
        '--last', dest='timestamp_last', nargs='?', default=None,
        help=(
            'Forces the last timestamp, and increments'
            'the preceding timestamps with the same amount.'
        )
    )
    parser.add_argument(
        '--interval', dest='timestamp_interval', nargs='?', default=None,
        help=(
            'Ignores the input timestamps and writes data with fixed timestamp intervals.\n'
            "Requires '--first'."
        )
    )
    parser.add_argument(
        '--offset', dest='timestamp_offset', nargs='?', default=None,
        help=(
            'A time interval to add to the input timestamp.'
        )
    )


def handle_series_cmd(client: WaylayClient, args: Namespace) -> bool:
    """Execute a 'series' command."""
    if args.series_cmd == CMD_EXPORT:
        handle_series_export_cmd(client, args)
        return True
    elif args.series_cmd == CMD_IMPORT:
        handle_series_import_cmd(client, args)
        return True
    return False


def convert_metrics(metrics_arg: List[str]) -> List[Metric]:
    """Convert a metrics list argument."""
    try:
        return [
            Metric(name=parts[0], key=(parts[1] if len(parts) > 1 else parts[0]))
            for m in metrics_arg
            for parts in [m.split('=')]
        ]
    except ValueError as exc:
        raise ValueError(f'Invalid metrics specification: {metrics_arg}') from exc


def convert_resources(resources_arg: List[str]) -> List[Resource]:
    """Convert a resources list argument."""
    try:
        return [
            Resource(id=parts[0], key=(parts[1] if len(parts) > 1 else parts[0]))
            for r in resources_arg
            for parts in [r.split('=')]
        ]
    except ValueError as exc:
        raise ValueError(f'Invalid resources specification: {resources_arg}') from exc


@cli_exeption_decorator
def handle_series_export_cmd(client: WaylayClient, args: Namespace):
    """Execute the 'series export' command."""
    metrics = convert_metrics(args.metrics)
    resources = convert_resources(args.resources)
    column_names = create_column_arguments(args)
    if 'timestamp_column' not in column_names:
        column_names['timestamp_column'] = 'timestamp'

    series_settings = SeriesSettings(
        per_resource=args.per_resource,
        per_metric=args.per_metric,
        resources=resources,
        metrics=metrics,
        **column_names,
        timestamp_format=args.timestamp_format,
        timestamp_timezone=args.timestamp_timezone,
        timestamp_from=args.timestamp_from,
        timestamp_until=args.timestamp_until,
    )
    client.timeseries.etl_tool.export_series_as_csv(
        args.output,
        archive_type=args.archive_type,
        progress=not args.silent,
        settings=series_settings,
    )


@cli_exeption_decorator
def handle_series_import_cmd(client: WaylayClient, args: Namespace):
    """Execute the 'series import' command."""
    metrics = convert_metrics(args.metrics or []) or None
    resources = convert_resources(args.resources or []) or None
    column_names = create_column_arguments(args)

    series_settings = SeriesSettings(
        resources=resources,
        metrics=metrics,
        resource=args.default_resource,
        metric=args.default_metric,
        **column_names,
        timestamp_format=args.timestamp_format,
        timestamp_timezone=args.timestamp_timezone,
        timestamp_from=args.timestamp_from,
        timestamp_until=args.timestamp_until,
    )

    series_input: List[Union[io.TextIOBase, pathlib.Path]] = [
        TextIOPeekReader(reader=sys.stdin) if s == '-' else pathlib.Path(s)
        for s in args.series
    ]
    progress = not args.silent
    etl_import = client.timeseries.etl_tool.prepare_import(
        *series_input,
        settings=series_settings,
        name=args.name,
        temp_dir=args.temp_dir,
        progress=progress
    )
    if args.no_import:
        print(etl_import.import_file.path)
        return

    client.timeseries.etl_tool.initiate_import(
        etl_import,
        resource_update_level=args.resource_update_level,
        progress=progress
    )


COLUMN_NAMES = ('timestamp', 'metric', 'resource', 'value')


def create_column_arguments(args: Namespace):
    """Create a dictionary out of column arguments."""
    column_names: Dict[str, Any] = {}
    if args.columns:
        for col_spec in args.columns:
            if '=' not in col_spec:
                col_spec = f'{col_spec}={col_spec}'
            try:
                column, name = col_spec.split('=')
                if column not in COLUMN_NAMES:
                    raise ValueError(f'Invalid column specification: {col_spec}')
                column_names[f'{column}_column'] = name or None
            except ValueError as exc:
                raise ValueError(f'Invalid column specification: {col_spec}') from exc
    return column_names


class TextIOPeekReader(io.TextIOBase):
    """A reader around stdin (or other readonly stream) that allows to peek the first n lines."""

    def __init__(self, peek_limit: int = 2, reader: TextIO = sys.stdin):
        """Create a TextIOPeekReader."""
        super().__init__()
        self.reader = reader
        self.peek_reads: List[str] = []
        self.peek_limit = peek_limit
        self.line_count = 0

    def readline(self):
        """Read a line from underlying text stream, buffering the first n."""
        line_count = self.line_count
        self.line_count += 1
        if line_count <= self.peek_limit:
            if line_count < len(self.peek_reads):
                return self.peek_reads[line_count]
            line = self.reader.readline()
            self.peek_reads.append(line)
            return line
        return self.reader.readline()

    def seek(self, offset: int, whence: int = 0):
        """If only first lines are read, reset to read from buffer."""
        line_count = self.line_count
        self.line_count = 0
        if whence == 0 and offset == 0 and line_count <= self.peek_limit:
            return
        # seek above the peek limit, try (and probably fail) to seek the underlying stream
        self.reader.seek(offset, whence)
        self.peek_reads = []



"""REST client for the Waylay Resources API Service."""

from waylay.service import WaylayRESTService

from .resource import ResourceResource
from .resource_type import ResourceTypeResource


class ResourcesService(WaylayRESTService):
    """REST client for the main Waylay Resources Service."""

    config_key = 'resources'
    service_key = 'resources'
    gateway_root_path = '/resources/v1'
    resource_definitions = {
        'resource': ResourceResource,
        'resource_type': ResourceTypeResource
    }
    resource: ResourceResource
    resource_type: ResourceTypeResource



"""REST client for the Waylay Storage Service."""

from waylay.service import WaylayRESTService

from .bucket import BucketResource
from .object import ObjectResource, FolderResource
from .content import ContentTool
from .about import AboutResource
from .subscription import SubscriptionResource


class StorageService(WaylayRESTService):
    """REST client for the Waylay Storage Service."""

    config_key = 'storage'
    service_key = 'storage'
    gateway_root_path = '/storage/v1'

    resource_definitions = {
        'bucket': BucketResource,
        'object': ObjectResource,
        'folder': FolderResource,
        'subscription': SubscriptionResource,
        'about': AboutResource,
        'content': ContentTool,
    }

    bucket: BucketResource
    object: ObjectResource
    folder: FolderResource
    subscription: SubscriptionResource
    about: AboutResource
    content: ContentTool



"""Utilities to parse and render ETL timeseries files."""
from typing import (
    Optional, Any
)
import gzip
import csv
from dataclasses import replace

import pandas as pd
from tqdm.std import tqdm

from waylay.exceptions import RequestError

from .model import (
    Metric,
    Resource,
    SeriesSettings,
    ETLFile,
    WaylayETLSeriesImport,

    SeriesInput,
    SeriesIterator,
    ETL_IMPORT_COLUMN_NAMES,
    render_timestamp_zulu,
    ParserRequestError,
)
from .import_csv import (
    MeasurementsStoreType,
    prepare_settings_csv,
    create_import_series_provider
)
from .import_dataframe import (
    prepare_settings_dataframe,
    iter_timeseries_dataframe
)
from .etlfile import (
    read_etl_import_as_stream,
    read_etl_import,
    dataframe_from_iterator,
    list_resources,
)

from .export_series import (
    export_csv,
    export_csv_to_tar_archive,
    export_csv_to_zip_archive,
    export_csv_to_dir,
    create_export_series_provider,
)


def iter_timeseries(
    *input_data: SeriesInput,
    settings: SeriesSettings,
    progress: bool = False,
    measurements_store: MeasurementsStoreType = MeasurementsStoreType.SPOOLED
) -> SeriesIterator:
    """Create a SeriesIterator for the given input source and settings."""
    if not input_data:
        raise AttributeError('input_data required.')
    if isinstance(input_data[0], pd.DataFrame):
        yield from iter_timeseries_dataframe(*input_data, settings=settings, progress=progress)
    else:
        with create_import_series_provider(
            *input_data, settings=settings, progress=progress, measurements_store=measurements_store
        ) as series_provider:
            if progress:
                with tqdm(
                    total=series_provider.row_count, unit='rows',
                    unit_scale=True, unit_divisor=1
                ) as tqdm_progress:
                    for (resource_id, metric_name), measurements in series_provider.items():
                        tqdm_progress.update()
                        yield resource_id, metric_name, iter(measurements)
            else:
                for (resource_id, metric_name), measurements in series_provider.items():
                    yield resource_id, metric_name, iter(measurements)


def prepare_etl_import(
    *series_input: SeriesInput,
    import_file: Optional[ETLFile] = None,
    settings: Optional[SeriesSettings] = None,
    **settings_args: Any
) -> WaylayETLSeriesImport:
    """Update the mapping settings by extracting metadata from an input source."""
    import_file = import_file or ETLFile()
    settings = settings or SeriesSettings()
    if settings_args:
        settings = replace(settings, **settings_args)

    # input settings validation
    etl_import = WaylayETLSeriesImport(
        series_input=series_input,
        settings=settings,
        import_file=import_file
    )

    # input series validation
    if not series_input:
        raise AttributeError('series_input required.')
    if isinstance(series_input[0], pd.DataFrame):
        if len(series_input) > 1:
            raise AttributeError('Multiple dataframe series_input not yet supported.')
        settings = prepare_settings_dataframe(series_input[0], settings=settings)
        etl_import = replace(etl_import, settings=settings)
    else:
        settings = prepare_settings_csv(*series_input, settings=settings)
        etl_import = replace(etl_import, settings=settings)
    return etl_import


def create_etl_import(
    etl_import: WaylayETLSeriesImport,
    progress: bool = True,
    measurements_store: MeasurementsStoreType = MeasurementsStoreType.SPOOLED
) -> WaylayETLSeriesImport:
    """Create an ETL import file from the given input."""
    file_path = etl_import.import_file.path
    if file_path.exists():
        raise RequestError(
            f'The file {file_path} already exists. '
            'Please remove or specify another etl-import file name.'
        )

    timeseries_data = iter_timeseries(
        *etl_import.series_input,
        settings=etl_import.settings,
        progress=progress,
        measurements_store=measurements_store
    )
    with gzip.open(file_path, 'wt') as csv_file:
        writer = csv.writer(csv_file, delimiter=',', quoting=csv.QUOTE_MINIMAL)
        writer.writerow(ETL_IMPORT_COLUMN_NAMES)
        for resource, metric, measurements in timeseries_data:
            if not etl_import.settings.skip_empty_values:
                measurements_it = (
                    (t, v) for t, v in measurements if t is not None
                )
            else:
                measurements_it = (
                    (t, v) for t, v in measurements if t is not None and v is not None and v != ''
                )

            for timestamp, value in measurements_it:
                writer.writerow([
                    resource,
                    metric,
                    render_timestamp_zulu(timestamp),
                    value
                ])
    return etl_import



"""REST definitions for the 'resource_type' entity of the 'api' service."""

from waylay.service import WaylayResource
from waylay.service import decorators

DEFAULT_DECORATORS = [decorators.exception_decorator, decorators.return_body_decorator]

RESOURCE_TYPE_ID_ARG = {
    'name': 'resource_type_id',
    'type': 'str',
    'description': 'id of a Waylay resource type.',
    'example': 'CustomSensor001'
}
RESOURCE_TYPE_BODY_ARG = {
    'name': 'body',
    'type': 'dict',
    'description': 'Representation of a resource type entity',
    'examples': ["""{
            "name" : "CustomSensor001",
            "id": "dvtp_custom_sensor_001"
        }"""]
}
ADDITIONAL_PARAMS_ARG = {
    'name': 'params',
    'type': 'dict',
    'description': 'Additional parameters, mapped to url query parameters. See API documentation.'
}


class ResourceTypeResource(WaylayResource):
    """REST Resource for the 'resource_type' entity of the 'api' resource provisioning service."""

    link_roots = {
        'doc': '${doc_url}/api/resources/?id=',
        'apidoc': '${apidoc_url}/resources.html'
    }

    actions = {
        'create': {
            'method': 'POST',
            'url': '/resourcetypes',
            'arguments': [RESOURCE_TYPE_BODY_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Create a `resource type` entity.',
            'links': {
                'doc': 'create-resource-type',
                'apidoc': '',
            },
        },
        'remove': {
            'method': 'DELETE',
            'url': '/resourcetypes/{}',
            'arguments': [RESOURCE_TYPE_ID_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Delete a `resource type` entity.',
            'links': {
                'doc': 'delete-resource-type',
                'apidoc': '',
            },
        },
        'replace': {
            'method': 'PUT',
            'url': '/resourcetypes/{}',
            'arguments': [RESOURCE_TYPE_ID_ARG, RESOURCE_TYPE_BODY_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Replace a `resource type` representation.',
            'links': {
                'doc': 'update-resource-type',
                'apidoc': '',
            },
        },
        'update': {
            'method': 'PATCH',
            'url': '/resourcetypes/{}',
            'arguments': [RESOURCE_TYPE_ID_ARG, RESOURCE_TYPE_BODY_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': '(Partially) update a `resource type` representation.',
            'links': {
                'doc': 'partial-resource-type-update',
                'apidoc': '',
            },
        },
        'get': {
            'method': 'GET',
            'url': '/resourcetypes/{}',
            'arguments': [RESOURCE_TYPE_ID_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Retrieve a `resource type` representation.',
            'links': {
                'doc': 'retrieve-resource-type',
                'apidoc': '',
            },
        },
        'list': {
            'method': 'GET',
            'url': '/resourcetypes',
            'arguments': [ADDITIONAL_PARAMS_ARG],
            'decorators': DEFAULT_DECORATORS,
            'description': 'Query `resource type` entities.',
            'links': {
                'doc': 'query-resource-types',
                'apidoc': '',
            },
        },
    }



"""REST definitions for version status of Waylay Queries Service."""

from .._base import WaylayResource
from .._decorators import (
    exception_decorator,
    return_path_decorator
)


class AboutResource(WaylayResource):
    """Version information."""

    actions = {
        'version': {
            'method': 'GET',
            'url': '/',
            'decorators': [exception_decorator, return_path_decorator(['version'])],
            'description': 'Version info of the <em>Queries Service</em> at this endpoint.'
        }
    }



"""waylay-beta build configuration"""
import json
from setuptools import setup, find_namespace_packages
import versioneer


with open("doc/byoml_runtimes.json", "r") as fh:
    runtimes_data = json.load(fh)

framework_extras = set()
runtime_extras = set()
runtime_requirements = {}
for runtime in runtimes_data['runtimes']:
    framework = runtime['framework']
    version = runtime['version']
    name = runtime.get('name', f'byoml-{framework}-{version}')
    runtime_extras.add(name)
    require = runtime.get('provided', [])
    runtime_requirements[name] = require
    framework_extra = f'byoml-{framework}'
    if framework_extra not in framework_extras:
        # first runtime is default for framework
        framework_extras.add(framework_extra)
        runtime_requirements[framework_extra] = require

with open("doc/dist.README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
    long_description = long_description.replace('$FRAMEWORK_EXTRA_LIST', ', '.join(framework_extras))
    long_description = long_description.replace('$RUNTIME_EXTRA_LIST', ', '.join(runtime_extras))


setup(
    name='waylay-beta',
    description='beta release of the Waylay Python SDK',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://docs.waylay.io/#/api/sdk/python',
    author='Waylay',
    author_email='info@waylay.io',
    license='ISC',
    license_file='LICENSE.txt',
    packages=find_namespace_packages(),
    package_data={"waylay": ["py.typed"]},
    include_package_data=True,
    install_requires=[
        'httpx',
        'simple-rest-client',
        'appdirs',
        'pyjwt',
        'numpy<2',
        'pandas<2',
        'isodate',
        'joblib',
        'tqdm',  # progress bar
        'tenacity',
        'tabulate',
    ],
    extras_require={
        'dev': [
            'pytest',
            'pytest-cov',
            'pytest-mock',
            'mock',
            'pylint',
            'pycodestyle',
            'pydocstyle',
            'autopep8',
            'mypy<0.990',
            'typing-inspect',
            'types-pytz',
            'types-setuptools',
            'types-tabulate',
            'pdoc',
        ],
        'dev-3.11': [
            'build',
        ],
        'dev-3.10': [
        ],
        ':python_version < "3.10"': [
            'importlib_metadata'
        ],
        **runtime_requirements
    },
    setup_requires=[
        'setuptools-pep8'
    ],
    python_requires='>=3.7',
    entry_points={
        "console_scripts": [
            "waylaycli = waylay.cli.waylaycli:main"
        ],
        "waylay_services": [
            "beta = waylay.service:SERVICES"
        ]
    }
)



"""Utilitly package for working with Pandas dataframes."""
__docformat__ = "google"
from collections import OrderedDict
import datetime
import functools
import re

from typing import (
    Any, Optional, Dict
)
import isodate
import pandas as pd
import pandas.tseries.frequencies as pd_freq

KEY_DEFAULT = ''


@pd.api.extensions.register_index_accessor("tsa")
class TsaIndexMetaAccessor:
    """Accessor added under the 'tsa' namespace to Indexes.

    Attaches additional timeseries functionality.
    A 'freq' property is added which will be taken from the index if available
    """

    def __init__(self, index):
        """Create the accessor."""
        self._index = index
        self._iso_freqstr = to_iso_freq(index.freq) if isinstance(index, pd.DatetimeIndex) else None
        self._table_attributes = {}

    @property
    def iso_freqstr(self) -> Optional[str]:
        """Get the ISO-8601 frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        return self._iso_freqstr

    @property
    def pandas_freqstr(self) -> Optional[str]:
        """Get the pandas frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        iso_freqstr = self.iso_freqstr
        if iso_freqstr is not None:
            return to_pandas_freqstr(iso_freqstr)
        return None

    @property
    def pandas_freq(self) -> Optional[pd_freq.DateOffset]:
        """Get the pandas frequency object for the DataFrame's index.

        None if there is no frequency.
        """
        iso_freqstr = self.iso_freqstr
        if iso_freqstr is not None:
            return to_pandas_freq(iso_freqstr)
        return None

    @property
    def window_length(self) -> Optional[pd.Timedelta]:
        """Get window length of this data series.

        If a frequency is present, this is the
        length of the half-open interval that extends
        1 freq unit after the last datapoint.
        """
        idx = self._index
        if isinstance(idx, pd.DatetimeIndex):
            if idx.size == 0:
                return None
            freq = self.pandas_freq
            if idx.size == 1 and freq is None:
                return None
            if freq is None:
                return idx[-1] - idx[0]
            else:
                return idx[-1] + freq - idx[0]
        return None

    @property
    def _from(self) -> Optional[pd.Timestamp]:
        """Get the `from` specification of this index."""
        idx = self._index
        if isinstance(idx, pd.DatetimeIndex) and idx.size > 0:
            return idx[0]
        return None

    @property
    def until(self) -> Optional[pd.Timestamp]:
        """Get the `until` specification of this index."""
        idx = self._index
        if isinstance(idx, pd.DatetimeIndex) and idx.size > 0:
            freq = self.pandas_freq
            if freq is not None:
                return idx[-1] + freq
            else:
                return idx[-1]
        return None

    @property
    def window_spec(self) -> Optional[Dict]:
        """Create the window specification that can be inferred."""
        idx = self._index
        if isinstance(idx, pd.DatetimeIndex) and idx.size > 0:
            spec = {
                "from": self._from,
                "until": self.until
            }
            window = self.window_length
            if window:
                spec["window"] = window
            freq = self.iso_freqstr
            if freq:
                spec["freq"] = freq
            return spec
        else:
            return None


@pd.api.extensions.register_series_accessor("tsa")
class TsaSeriesMetaAccessor:
    """Accessor added under the 'tsa' namespace to Series.

    Attaches additional timeseries functionality.
    """

    def __init__(self, series: pd.Series):
        """Create a TsaSeriesMetaAccessor."""
        self._series = series
        self._series_attributes: Dict[str, Any] = {}

    @property
    def table_attributes(self) -> Dict:
        """Get additional attributes on table level.

        E.g. column metadata that is the same for each column.
        """
        return self._series_attributes

    @property
    def iso_freqstr(self) -> Optional[str]:
        """Get the ISO-8601 frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        return self._series.index.tsa.iso_freqstr

    @property
    def pandas_freqstr(self) -> Optional[str]:
        """Get the pandas frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        return self._series.index.tsa.pandas_freqstr

    @property
    def pandas_freq(self) -> Optional[pd_freq.DateOffset]:
        """Get the pandas frequency object for the DataFrame's index.

        None if there is no frequency.
        """
        return self._series.index.tsa.pandas_freq

    @property
    def window_length(self) -> Optional[pd.Timedelta]:
        """Get window length of this data series."""
        return self._series.index.tsa.window_length

    @property
    def _from(self) -> Optional[pd.Timestamp]:
        """Get the `from` specification of this Series."""
        return self._series.index.tsa._from  # pylint: disable=protected-access

    @property
    def until(self) -> Optional[pd.Timestamp]:
        """Get the `until` specification of this Series."""
        return self._series.index.tsa.until

    @property
    def window_spec(self) -> Dict:
        """Get the window specification of this Series."""
        return self._series.index.tsa.window_spec


@pd.api.extensions.register_dataframe_accessor("tsa")
class TsaDataframeMetaAccessor:
    """TSA accessor added under the 'tsa' namespace to Dataframes.

    Attaches additional timeseries functionality.
    """

    def __init__(self, df):
        """Create a TsaDataframeMetaAccessor."""
        self._df = df
        self._table_attributes = OrderedDict()

    @property
    def table_attributes(self):
        """Additional attributes on table level.

        e.g. column metadata that is the same for each column.
        """
        return self._table_attributes

    @property
    def iso_freqstr(self) -> Optional[str]:
        """Get the ISO-8601 frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        return self._df.index.tsa.iso_freqstr

    @property
    def pandas_freqstr(self) -> Optional[str]:
        """Get the pandas frequency string for the DataFrame's index.

        None if there is no frequency.
        """
        return self._df.index.tsa.pandas_freqstr

    @property
    def pandas_freq(self) -> Optional[pd_freq.DateOffset]:
        """Get the pandas frequency object for the DataFrame's index.

        None if there is no frequency.
        """
        return self._df.index.tsa.pandas_freq

    @property
    def window_length(self) -> Optional[str]:
        """Get window length of this DataFrame as an iso 8601 period string."""
        return self._df.index.tsa.window_length

    @property
    def _from(self):
        """Get the `from` specification of this DataFrame."""
        return self._df.index.tsa._from  # pylint: disable=protected-access

    @property
    def until(self):
        """Get the `until` specification of this DataFrame."""
        return self._df.index.tsa.until

    @property
    def window_spec(self):
        """Get the window_spec specification of this DataFrame."""
        return self._df.index.tsa.window_spec  # pylint: disable=protected-access


def to_pandas_freq(freq: Any) -> pd_freq.DateOffset:
    """Convert a frequency or interval to a pandas DateOffset."""
    if freq is None:
        return None
    if isinstance(freq, str) and 'P' in freq:
        freq = isodate.parse_duration(freq)
    if isinstance(freq, isodate.Duration):
        # If we've got a duration, we need to root it somewhere to
        # convert it into a frequency.
        # Best would be to root it at the context time,
        # but that isn't always available, so we just use the unix epoch
        # to have a consistent answer for the same question.
        freq = freq.totimedelta(start=datetime.datetime(1970, 1, 1, 0, 0, 0))
    return pd_freq.to_offset(freq)


def to_pandas_freqstr(freq: Any) -> Optional[str]:
    """Convert a frequency or interval to a pandas frequency string."""
    if freq is None:
        return None
    return to_pandas_freq(freq).freqstr


def _iso_format(postfix, prefix='P', scale=1, value=1):
    if scale != 1:
        value = int(value) * scale
    return "{}{}{}".format(prefix, value, postfix)


def _iso_formatter(postfix, prefix='P', scale=1):
    return functools.partial(_iso_format, postfix, prefix=prefix, scale=scale)


ISO_PERIOD_FORMAT_BY_PANDAS_UNIT = {
    'B': _iso_formatter('D'),
    'C': _iso_formatter('D'),
    'D': _iso_formatter('D'),

    'W': _iso_formatter('W'),

    'M': _iso_formatter('M'),
    'BM': _iso_formatter('M'),
    'CBM': _iso_formatter('M'),
    'MS': _iso_formatter('M'),
    'BMS': _iso_formatter('M'),
    'CBMS': _iso_formatter('M'),

    'SM': _iso_formatter('M', scale=0.5),
    'SMS': _iso_formatter('M', scale=0.5),

    'Q': _iso_formatter('M', scale=3),
    'BQ': _iso_formatter('M', scale=3),
    'QS': _iso_formatter('M', scale=3),
    'BQS': _iso_formatter('M', scale=3),

    'A': _iso_formatter('Y'),
    'Y': _iso_formatter('Y'),
    'BA': _iso_formatter('Y'),
    'BY': _iso_formatter('Y'),
    'AS': _iso_formatter('Y'),
    'YS': _iso_formatter('Y'),
    'BAS': _iso_formatter('Y'),
    'BYS': _iso_formatter('Y'),

    'BH': _iso_formatter('H', 'PT'),
    'H': _iso_formatter('H', 'PT'),

    'T': _iso_formatter('M', 'PT'),
    'min': _iso_formatter('M', 'PT'),

    'S': _iso_formatter('S', 'PT'),

    'L': _iso_formatter('S', 'PT', 0.001),
    'ms': _iso_formatter('S', 'PT', 0.001),
    'U': _iso_formatter('S', 'PT', 0.000001),
    'us': _iso_formatter('S', 'PT', 0.000001),
    'N': _iso_formatter('S', 'PT', 0.000000001),
}


def to_iso_freq(freq, lsnr=None) -> Optional[str]:
    """Convert a frequency or interval to an ISO-8601 Duration."""
    if freq is None:
        return None
    if isinstance(freq, str) and 'P' in freq:
        return freq
    freq = to_pandas_freqstr(freq)
    # didnt find a good way, so like this:
    if '-' in freq:
        # remove anchoring part
        freq = freq.split('-')[0]
    m = re.search(r'(\d+)?([BCDWMSQAYHTminusLUN]+)', freq)
    if m is None:
        if lsnr is not None:
            lsnr.warning("could not convert frequency to isoformat: %s", freq)
        return freq
    freq_n = m.group(1)
    freq_n = '1' if freq_n is None else freq_n
    freq_unit = m.group(2)

    if freq_unit in ISO_PERIOD_FORMAT_BY_PANDAS_UNIT:
        return ISO_PERIOD_FORMAT_BY_PANDAS_UNIT[freq_unit](value=freq_n)
    else:
        if lsnr is not None:
            lsnr.info("could not convert frequency to isoformat: %s", freq)
        return freq



"""REST definitions for the 'bucket' entity of the 'storage' service."""
from urllib.parse import quote

from waylay.service import WaylayResource, decorators

BUCKET_NAME_ARG = {
    'name': 'bucket_name',
    'type': 'str',
    'description': 'Name of a Waylay storage bucket.',
    'examples': ['assets', 'public']
}
ADDITIONAL_PARAMS_ARG = {
    'name': 'params',
    'type': 'dict',
    'description': 'Additional parameters, mapped to url query parameters. See API documentation.'
}


class BucketResource(WaylayResource):
    """REST Resource for the 'bucket' entity of the 'storage' service."""

    link_roots = {
        'doc': '${doc_url}/api/storage/?id=',
        'apidoc': '${apidoc_url}/storage.html'
    }

    actions = {
        'list': {
            'method': 'GET', 'url': '/bucket', 'decorators': [
                decorators.exception_decorator,
                decorators.return_path_decorator(['buckets']),
            ],
            'description': 'List available bucket aliases',
            'links': {
                'doc': 'list-bucket',
                'apidoc': '',
            }
        },
        'get': {
            'method': 'GET', 'url': '/bucket/{}', 'decorators': [
                decorators.exception_decorator,
                decorators.return_body_decorator,
            ],
            'arguments': [BUCKET_NAME_ARG],
            'description': 'Get metadata for a specific bucket alias',
            'links': {
                'doc': 'get-bucket',
                'apidoc': '',
            }
        },
    }



"""Exceptions specific to the Queries Service."""

from typing import (
    List, Mapping, Any
)
from ...exceptions import RestRequestError, RestResponseError, RestResponseParseError


class QueryActionError(RestResponseError):
    """Error that represents the json messages of a query response."""

    @property
    def messages(self) -> List[Mapping[str, Any]]:
        """Get the list of message objects returned by a query error response."""
        return self._get_from_body('messages', [])

    @property
    def message(self):
        """Get the main user error returned by a query error response."""
        return self._get_from_body('error', super().message)


class QueryActionParseError(RestResponseParseError, QueryActionError):
    """Indicates that a query response could not be parsed."""


class QueryRequestError(RestRequestError):
    """Indicates issues with the request arguments."""



"""REST definitions for handling content from the 'storage' service objects."""
__docformat__ = "google"

from typing import Optional, Any, Union, IO, List, Dict
from pathlib import Path
import os

import httpx
import httpx._types as _http_types

from tqdm import tqdm

from waylay.exceptions import RestResponseError, RestRequestError
from waylay.service import WaylayResource, WaylayRESTService


_http = httpx


class ContentTool(WaylayResource):
    """Tool for handling content from of the 'storage' service objects using signed get and put links."""

    service: WaylayRESTService

    actions: Dict[str, Any] = {
        'put': {
            'wrapped_actions': []
        },
        'get': {
            'wrapped_actions': []
        }
    }

    def put(
        self,
        bucket: str,
        path: str,
        content_type: Optional[str] = None,
        content: Optional[_http_types.RequestContent] = None,
        json: Any = None,
        from_file: Optional[Union[str, os.PathLike]] = None,
        progress: bool = True,
        headers: Optional[_http_types.HeaderTypes] = None,
        raw: bool = False,
        **kwargs
    ) -> Union[_http.Response, str]:
        """Store the content of a storage object.

Content can be specified with (exactly) one of the attributes 'from_file', 'content' or 'json'.

This method gets a signed url (see storage.object.sign_put),
and uploads the object content to the object store.

Arguments:
    bucket          The name of the bucket.
    path            The object path for the uploaded content.
    content_type    [optional] The content type with which this content will be available.
    content         A byte array or byte stream. With streaming data, you will need
                    to provide a 'content-length' header.
    json            A data structure as to be uploaded with JSON encoding.
    from_file       A file path of the file content that needs to be uploaded.
    progress        [optional, bool] If true, the client will show a progress bar.
    headers         [optional] Additional headers to be used for the upload.
    raw             [optional, bool] If True, the HTTP response object is returned,
                    even in case of failures. Otherwise, return the status text or
                    raises errors on failure.
    kwargs          Additional attributes forwarded to the 'sign_get' request.
        """
        params = kwargs.pop('params', {})
        params.update(kwargs)
        headers = _http.Headers(headers)
        if content_type:
            params['content_type'] = content_type
            headers.update({'content-type': content_type})

        upload_url = self.service.object.sign_put(bucket, path, params=params)

        if sum(input is not None for input in [content, json, from_file]) != 1:
            raise RestRequestError(
                "Only one of the 'content', 'json' or 'from_file' arguments"
                " should be specified."
            )

        if progress:
            print(
                f'Uploading content to {bucket}/{path} ...'
            )
        if from_file:
            file_path = Path(from_file)
            headers.update({'content-length': format(file_path.stat().st_size)})
            with open(from_file, 'rb') as file_content:
                response = _http.put(
                    upload_url,
                    content=file_content,
                    headers=headers
                )
        else:
            response = _http.put(
                upload_url,
                content=content,
                json=json,
                headers=headers
            )

        if progress:
            print('... done.')

        if raw:
            return response

        self._raise_for_status(f'Failed to upload for {bucket}/{path}', response)
        return response.reason_phrase

    def _write_response_to_file_with_progress(self, response: _http.Response, file_stream: IO):
        total = int(response.headers['Content-Length'])
        with tqdm(total=total, unit_scale=True, unit_divisor=1024, unit="B") as progress:
            num_bytes_downloaded = response.num_bytes_downloaded
            for chunk in response.iter_bytes():
                file_stream.write(chunk)
                progress.update(response.num_bytes_downloaded - num_bytes_downloaded)
                num_bytes_downloaded = response.num_bytes_downloaded

    def _write_response_to_file(self, response: _http.Response, file_stream: IO):
        for chunk in response.iter_bytes():
            file_stream.write(chunk)

    def _raise_for_status(self, message: str, response: _http.Response):
        try:
            response.raise_for_status()
        except _http.HTTPStatusError as exc:
            raise RestResponseError(message, exc.response) from exc

    def _get_to_file(self, get_url: str, to_file: Path, progress: bool):
        with open(to_file, 'wb') as download_stream:
            with _http.stream('GET', get_url) as response:
                if response.status_code >= 300:
                    return response
                if progress:
                    self._write_response_to_file_with_progress(response, download_stream)
                else:
                    self._write_response_to_file(response, download_stream)
                return response

    def get(
        self,
        bucket: str,
        path: str,
        to_file: Optional[Union[str, os.PathLike]] = None,
        progress: bool = True,
        headers: Optional[_http_types.HeaderTypes] = None,
        raw: bool = False,
        **kwargs
    ) -> Union[_http.Response, Path]:
        """Retrieve the content of a storage object.

This method gets the signed_url (see storage.object.sign_get),
and downloads the object content from the given url to the object store.

Arguments:
    bucket          The name of the bucket.
    path            The object path for the content to be downloaded.
    to_file         A local file path to which the content should be saved.
                    If succesfull, will return a 'Path' object to that file.
    progress        [optional, bool] If true, the client will show a progress bar.
    headers         [optional] Additional headers to be used for the download.
    raw             [optional, bool] If 'True', the HTTP response object is returned,
                    even in case of failures. Otherwise, raises errors on failure.

    kwargs          Additional attributes forwarded to the 'sign_get' request.

Returns:
    If 'to_file' is specified: a 'Path' object representing the download location.
    Otherwise, w response object having having a properties like
        'content', 'text', 'status_code', 'headers'
    and a
        'json()'
    method to extract the data.
        """
        params = kwargs.pop('params', {})
        params.update(kwargs)
        get_url = self.service.object.sign_get(bucket, path, params=params)

        if to_file:
            response = self._get_to_file(get_url, Path(to_file), progress)
        else:
            response = _http.get(get_url, headers=headers)

        if raw:
            return response

        self._raise_for_status(f'Failed to retrieve content for {bucket}/{path}', response)

        if to_file:
            return Path(to_file).absolute()

        return response
